"""
snowmobile.core.Configuration is a parsed snowmobile.toml file; class handles:

1.  Locating `snowmobile.toml`, from:

    a.  A cached location specific to the version of :xref:`snowmobile`
        and the file name (defaults to `snowmobile.toml`)
    b.  Finding a file based on its name from traversing the file system,
        used when initially finding snowmobile.toml or when a bespoke
        configuration file name has been provided
        
2.  Checking **[ext-sources]** for specified external configurations
3.  Instantiating each section in snowmobile.toml from the (Pydantic)
    models defined in :mod:`snowmobile.core.cfg`; root sections are set as
    individual attributes on :class:`~snowmobile.core.configuration.Configuration`

"""
from __future__ import annotations

import json
import shutil
from pathlib import Path
from types import MethodType
from typing import Any, Callable, Dict, List, Optional

import toml
from pydantic.json import pydantic_encoder

from snowmobile.core import paths, cfg, utils
from snowmobile.core.utils.parsing import rmerge_dicts
from .base import Generic
from .cache import Cache


class Configuration(Generic):
    """
    A parsed `snowmobile.toml` file.
    """

    # -- Statement components to be considered for scope.
    _SCOPE_ATTRIBUTES = ["kw", "obj", "desc", "anchor", "nm"]
    _SCOPE_TYPES = ["incl", "excl"]

    # -- Anchors to associate with QA st.
    QA_ANCHORS = {"qa-diff", "qa-empty"}

    def __init__(
        self,
        creds: Optional[str] = None,
        config_file_nm: Optional[str] = None,
        from_config: Optional[Path, str] = None,
        export_dir: Optional[Path, str] = None,
        silence: bool = False,
    ):
        """*All keyword arguments optional.*
        
        Args:
            config_file_nm (Optional[str]):
                Name of configuration file to use; defaults to `snowmobile.toml`.
            creds (Optional[str]):
                Alias for the set of credentials to authenticate with; default
                behavior will fall back to the `connection.default-creds`
                specified in `snowmobile.toml`, `or the first set of credentials
                stored if this configuration option is left blank`.
            from_config (Optional[str, Path]):
                A full path to a specific configuration file to use; bypasses
                any checks for a cached file location and can be useful for
                container-based processes with restricted access to the local
                file system.
            export_dir(Optional[Path]):
                Path to save a template `snowmobile.toml` file to; if pr,
                the file will be exported within the __init__ method and nothing
                else will be instantiated.
                
        """
        # fmt: off
        super().__init__()

        self._stdout = self.Stdout(silence=silence)
        """Stdout: Console output."""

        self.file_nm = config_file_nm or "snowmobile.toml"
        """str: Configuration file name; defaults to 'snowmobile.toml'."""

        self.cache = Cache()
        """snowmobile.core.cache.Cache: Persistent cache; caches :attr:`location`."""
        
        self.location = Path()
        """pathlib.Path: Full path to configuration file."""
        
        self.connection: Optional[cfg.Connection] = None
        """snowmobile.core.cfg.Connection: **[connection]** from snowmobile.toml."""
        
        self.loading: Optional[cfg.Loading] = None
        """snowmobile.core.cfg.Loading: **[loading]** from snowmobile.toml."""
        
        self.script: Optional[cfg.Script] = None
        """snowmobile.core.cfg.Script: **[script]** from snowmobile.toml."""
        
        self.sql: Optional[cfg.SQL] = None
        """snowmobile.core.cfg.SQL: **[sql]** from snowmobile-ext.toml."""
        
        self.ext_sources: Optional[cfg.Location] = None
        """snowmobile.core.cfg.Location: **[external-sources]** from snowmobile.toml."""

        # for snowmobile_template.toml export only
        if export_dir:
            self._stdout.exporting(file_name=self.file_nm)
            export_dir = export_dir or Path.cwd()
            export_path = export_dir / self.file_nm
            template_path = paths.DIR_PKG_DATA / "snowmobile-template.toml"
            if export_path.exists():
                raise FileExistsError(
                    f"Cannot overwrite existing `snowmobile.toml` at:\n\t{template_path}."
                )
            shutil.copy(template_path, export_path)
            self._stdout._exported(file_path=export_path)

        # standard case
        else:

            self.location: Path = (
                Path(str(from_config))
                if from_config
                else Path(str(self.cache.get(self.file_nm)))
            )

            try:

                # import snowmobile.toml
                path_to_config = self._get_path(is_provided=bool(from_config))
                with open(path_to_config, "r") as r:
                    cfg_raw = toml.load(r)

                # set 'provided-creds' value if alias is passed explicitly
                cfg_raw['connection']['pr-creds'] = creds.lower() if creds else ""

                # check for specified 'external-sources', else set to defaults
                ext_sources = cfg_raw["external-sources"]
                for src in [
                    ("snowmobile_ext", paths.EXTENSIONS_DEFAULT_PATH),
                    ("ddl", paths.DDL_DEFAULT_PATH),
                    ("sql-save-heading", paths.SQL_EXPORT_HEADING_DEFAULT_PATH),
                ]:
                    if not ext_sources.get(src[0]):
                        ext_sources[src[0]] = src[1]

                # import snowmobile-ext.toml
                path_ext = ext_sources['snowmobile_ext']
                with open(path_ext, 'r') as r:
                    snowmobile_ext = toml.load(r)

                # recursively merge snowmobile.toml and snowmobile-ext.toml -
                # note that the dict order matters; if d1 and d2 share a key at
                # the same level, the value from d1 will supersede that of d2
                merged = rmerge_dicts(
                    d1=cfg_raw,
                    d2=snowmobile_ext
                )

                # set root classes as configuration attributes
                # merged = self._find_and_merge_configurations()
                self.connection = cfg.Connection(**merged.get('connection', {}))
                self.loading = cfg.Loading(**merged.get('loading', {}))
                self.script = cfg.Script(**merged.get('script', {}))
                self.sql = cfg.SQL(**merged.get('sql', {}))
                self.ext_sources = cfg.Location(**merged.get('external-sources', {}))

            except IOError as e:
                raise IOError(e)

            # fmt: on
            
    def _find_and_merge_configurations(self, **kwargs):
        pass

    @property
    def markdown(self) -> cfg.Markup:
        """Accessor for cfg.script.markdown."""
        return self.script.markup

    @property
    def attrs(self) -> cfg.Attributes:
        """Accessor for cfg.script.markdown.attributes."""
        return self.script.markup.attrs

    @property
    def wildcards(self) -> cfg.Wildcard:
        """Accessor for cfg.script.patterns.wildcards."""
        return self.script.patterns.wildcards

    # noinspection PyProtectedMember
    def _get_path(self, is_provided: bool = False):
        """Checks for cache existence and validates - traverses OS if not.

        Args:
            is_provided (bool):
                Indicates whether or not `from_config` is populated or if
                configuration file is intended to come from cache or from
                a bottom-up traversal of the file system if not yet cached.
        """
        self._stdout._locating_outcome(is_provided=is_provided, file=self.file_nm)
        if self.location.is_file():
            self._stdout._found(file_path=self.location, is_provided=is_provided)
            return self.location
        # else:
        self._stdout._not_found(creds_file_nm=self.file_nm)
        return self._find_creds()

    def _find_creds(self) -> Path:
        """Traverses file system from ground up looking for creds file."""

        found = None
        try:
            rents = [p for p in Path.cwd().parents]
            rents.insert(0, Path.cwd())
            for rent in rents:
                if list(rent.rglob(self.file_nm)):
                    found = list(rent.rglob(self.file_nm))[0]
                    break
            if found.is_file():
                self.location = found
                self.cache.save_item(item_name=self.file_nm, item_value=self.location)
                # TODO: Add a silence method that will shut this up if desired
                self._stdout._file_located(file_path=self.location)
                return self.location

        except FileNotFoundError as e:
            self._stdout._cannot_find(self.file_nm)
            raise FileNotFoundError(e)

    @staticmethod
    def batch_set_attrs(obj: Any, attrs: dict, to_none: bool = False):
        """Batch sets attributes on an object from a dictionary.

        Args:
            obj (Any):
                Object to set attributes on.
            attrs (dict):
                Dictionary containing attributes.
            to_none (bool):
                Set all of the object's attributes batching a key in `wrap`
                to `None`; defaults ot `False`.

        Returns (Any):
            Object post-setting attributes.

        """
        for k in set(vars(obj)).intersection(attrs):
            setattr(obj, k, None if to_none else attrs[k])
        return obj

    @staticmethod
    def attrs_from_obj(
        obj: Any, within: Optional[List[str]] = None
    ) -> Dict[str, MethodType]:
        """Utility to return attributes/properties from an object as a dictionary."""
        return {
            str(m): getattr(obj, m)
            for m in dir(obj)
            if (m in within if within else True)
            and not isinstance(getattr(obj, m), Callable)
        }

    @staticmethod
    def methods_from_obj(
        obj: Any, within: Optional[List[str]] = None
    ) -> Dict[str, MethodType]:
        """Returns callable components of an object as a dictionary."""
        return {
            str(m): getattr(obj, m)
            for m in dir(obj)
            if (m in within if within else True)
            and isinstance(getattr(obj, m), MethodType)
        }

    @property
    def scopes(self):
        """All combinations of scope type and scope attribute."""
        # TODO: Stick somewhere that makes sense
        return {
            f"{typ}_{attr}": set()
            for typ in self._SCOPE_TYPES
            for attr in self._SCOPE_ATTRIBUTES
        }

    def scopes_from_kwargs(self, only_populated: bool = False, **kwargs) -> Dict:
        """Turns *script.filter()* arguments into a valid set of kwargs for :class:`Scope`.

        Returns dictionary of all combinations of 'arg' ("kw", "obj", "desc",
        "anchor" and "nm"), including empty sets for any 'arg' not included
        in the keyword arguments provided.

        """
        scopes = {}
        for attr in self.scopes:
            attr_value = kwargs.get(attr) or set()
            scopes[attr] = attr_value
        return {k: v for k, v in scopes.items() if v} if only_populated else scopes

    def scopes_from_tag(self, t: Any):
        """Generates list of keyword arguments to instantiate all scopes for a wrap."""
        return [
            {
                "base": self.methods_from_obj(obj=t, within=self._SCOPE_ATTRIBUTES)[k](),
                "arg": k
            }
            for k in self._SCOPE_ATTRIBUTES
        ]

    def json(self, by_alias: bool = False, **kwargs):
        """Serialization method for core object model."""
        total = {}
        for k, v in vars(self).items():
            if issubclass(type(v), cfg.Base):
                total = {**total, **v.as_serializable(by_alias=by_alias)}
        return json.dumps(obj=total, default=pydantic_encoder, **kwargs)

    def __json__(self, by_alias: bool = False, **kwargs):
        return self.json(by_alias=by_alias, **kwargs)

    def __str__(self):
        return f"snowmobile.Configuration('{self.file_nm}')"

    def __repr__(self):
        return f"snowmobile.Configuration(config_file_nm='{self.file_nm}')"

    # noinspection PyMissingOrEmptyDocstring,PyMethodMayBeStatic
    class Stdout(utils.Console):
        """Console output."""

        def __init__(self, silence: bool = False):
            super().__init__(silence=silence)

        def exporting(self, file_name: str):
            self.p(f"Exporting {file_name}..")

        def _exported(self, file_path: Path):
            path = self.offset_path(file_path=file_path)
            self.p(f"<1 of 1> Exported {path}")

        def _locating(self):
            self.p("Locating credentials..")
            return self

        def _checking_cache(self, file: str):
            self.p(f"(1 of 2) Finding {file}..")
            return self

        def _reading_provided(self):
            self.p("(1 of 2) Checking pr path...")
            return self

        def _locating_outcome(self, is_provided: bool, file: str):
            _ = self._locating()
            return self._checking_cache(file) if not is_provided else self._reading_provided()

        def _cache_found(self, file_path: Path):
            path = self.offset_path(file_path=file_path)
            self.p(f"(2 of 2) Cached path found at {path}")
            return self

        def _provided_found(self, file_path: Path):
            path = self.offset_path(file_path=file_path)
            self.p(f"(2 of 2) Reading pr path {path}")
            return self

        def _found(self, file_path: Path, is_provided: bool):
            return (
                self._cache_found(file_path)
                if not is_provided
                else self._provided_found(file_path)
            )

        def _cache_not_found(self):
            self.p("(2 of 2) Cached path not found")
            return self

        def _traversing_for(self, creds_file_nm: str):
            self.p(f"\nLooking for {creds_file_nm} in local file system..")
            return self

        def _not_found(self, creds_file_nm: str):
            return self._cache_not_found()._traversing_for(creds_file_nm=creds_file_nm)

        def _file_located(self, file_path: Path):
            path = self.offset_path(file_path=file_path)
            self.p(f"(1 of 1) Located '{file_path.name}' at {path}")
            return self

        def _cannot_find(self, creds_file_nm: str):
            self.p(
                f"(1 of 1) Could not find config file {creds_file_nm} - "
                f"please double check the name of your configuration file or "
                f"value passed in the 'creds_file_nm' argument"
            )
            return self



"""
:class:`SQL` contains utility methods to generate & execute
common SQL commands; :class:`~snowmobile.core.connection.Snowmobile` inherits
everything from this object and provides it with the .query() method for
statement execution.

.. note::
   The :attr:`~SQL.auto_run` attribute defaults to `True`, meaning that the
   ge sql will execute when a method is called; if set to `False`
   the method will return the sql as a string without executing.

   The :class:`SQL` object is primarily interacted with as a
   pre-instantiated attribute of :class:`~snowmobile.Snowmobile`; in these
   instances users can fetch the ge sql as a string either by:

   1. Providing *run=False* to any method called; this will override
      all behavior set by the current value of :attr:`auto_run`
   2. Setting the :attr:`auto_run` attribute to `False` on an existing
      instance of :class:`SQL`, which will replicate the behavior of
      `(1)` without needing to provide *run=False* to each method
      called on that instance

"""
from __future__ import annotations

import json
from pathlib import Path
from typing import Dict, List, Optional, Union, Callable

import pandas as pd

from . import Generic
from .configuration import Configuration
from .utils.parsing import p, up, strip as s
from .errors import SnowflakeObjectNotFound


class SQL(Generic):
    """SQL class for generation & execution of common sql commands.

    Intended to be interacted with as a parent of :class:`~snowmobile.Snowmobile`.

    .. note::
        *   All arguments except for :attr:`sn` are optional.
        *   The benefit of setting the other attributes on an instance of :class:`SQL`
            is to (optionally) avoid passing the same information to multiple methods
            when generating a variety of statements around the same object.

    Attributes:
        nm (str):
            Object name to use in ge sql (e.g. 'some_table_name')
        obj (str):
            Object type to use in ge sql (e.g. 'table')
        schema (str):
            Schema to use when dot-prefixing sql; defaults to the schema with which the
            :attr:`sn` is connected to.
        auto_run (bool):
            Indicates whether to automatically execute the sql ge by a given
            method; defaults to *True*

    """

    def __init__(
        self,
        _query_func: Callable,
        _cfg: Configuration,
        nm: Optional[str] = None,
        schema: Optional[str] = None,
        obj: Optional[str] = None,
        auto_run: Optional[bool] = True,
    ):
        """Initializes a :class:`snowmobile.SQL` object."""
        super().__init__()
        
        _schema, self.nm = p(nm=nm)
        
        self.schema = schema or _schema or _cfg.connection.current.schema_name
        
        self.obj: str = obj or "table"
        self.auto_run: bool = auto_run
        
        self._query = _query_func
        self._cfg = _cfg
        
    # -- Information Schema ---------------------------------------------------
    
    def info_schema(
        self,
        loc: str,
        where: Optional[List[str]] = None,
        fields: Optional[List[str]] = None,
        order_by: Optional[List] = None,
        run: Optional[bool] = None,
    ) -> Union[str, pd.DataFrame]:
        """Generic case of selecting from information schema location."""
        _fields = self.fields(fields=fields)
        _order_by = self.order(by=order_by)
        _loc = f"information_schema.{loc}"
        _where = [w for w in where or list() if w]
        if _where:
            clauses = 'and'.join(f"\n\t{clause}" for clause in _where)
            _where = f"where\n{clauses}"
        sql = f"""
select
{_fields}
from {_loc}
{_where}
{_order_by}
"""
        _sql = s(sql, trailing=False, blanks=True)
        return self._query(_sql) if self(run) else _sql
    
    def table_info(
        self,
        nm: Optional[str] = None,
        fields: List[str] = None,
        restrictions: Dict[str, str] = None,
        order_by: List[Optional[str, int]] = None,
        all_schemas: bool = False,
        run: Optional[bool] = None,
    ) -> Union[str, pd.DataFrame]:
        """Query ``information_schema.tables`` for a given table or view.

        Args:
            nm (str):
                Table name, including schema if creating a stage outside of the
                current schema.
            fields (List[str]):
                List of fields to include in returned results (e.g.
                ['table_name', 'table_type', 'last_altered'])
            restrictions (List[str]):
                List of conditionals typed as literal components of a `where`
                clause (e.g.
                ["table_type = 'base table'", 'last_altered::date = current_date()']
                ).
            order_by (List[str]):
                List of fields or their ordinal positions to order the results by.
            all_schemas (bool):
                Include tables/views from all schemas; defaults to `False`.
            run (bool):
                Determines whether to run the ge sql or not; defaults to `None`
                which will reference the current value of the :attr:`auto_run` attribute
                which defaults to `True`.

        Returns (Union[str, pd.DataFrame]):
            Either:
                1.  The results of the query as a :class:`pandas.DataFrame`, or
                2.  The ge query as a :class:`str` of sql.

        """
        # fmt: off
        try:
            schema, nm = p(nm)
            table, schema = (
                self._validate(
                    val=(nm or self.nm), nm='nm', attr_nm='nm'
                ),
                self._validate(
                    val=(schema or self.schema), nm='schema', attr_nm='schema'
                )
            )

        except ValueError as e:
            raise e

        restrictions = {
            **(restrictions or dict()),
            **{
                "lower(table_name)": f"'{table.lower()}'",
                "lower(table_schema)": f"'{schema.lower()}'",
            },
        }
        if all_schemas:
            _ = restrictions.pop("lower(table_schema)")

        sql = self._info_schema_generic(
            obj="table", fields=fields, restrictions=restrictions, order_by=order_by
        )
        # fmt: on

        return self._query(sql=sql) if self(run) else sql

    def column_info(
        self,
        nm: Optional[str] = None,
        fields: Optional[List] = None,
        restrictions: Optional[Dict] = None,
        order_by: Optional[List] = None,
        all_schemas: bool = False,
        run: Optional[bool] = None,
    ) -> Union[str, pd.DataFrame]:
        """Query ``information_schema.columns`` for a given table or view.

        Args:
            nm (str):
                Table name, including schema if creating a stage outside of the
                current schema.
            fields (List[str]):
                List of fields to include in returned results (e.g.
                ['ordinal_position', 'column_name', 'data_type'])
            restrictions (List[str]):
                List of conditionals typed as literal components of a `where`
                clause (e.g.["regexp_count(lower(column_name), 'tmstmp') = 0"]).
            order_by (List[str]):
                List of fields or their ordinal positions to order the results by.
            all_schemas (bool):
                Include tables/views from all schemas; defaults to `False`.
            run (bool):
                Determines whether to run the ge sql or not; defaults to `None`
                which will reference the current value of the :attr:`auto_run` attribute
                which defaults to `True`.

        Returns (Union[str, pd.DataFrame]):
            Either:
                1.  The results of the query as a :class:`pandas.DataFrame`, or
                2.  The ge query as a :class:`str` of sql.

        """
        try:
            # fmt: off
            schema, nm = p(nm)
            table = self._validate(
                val=(nm or self.nm), nm='nm', attr_nm='nm'
            )
            schema = self._validate(
                val=(schema or self.schema), nm='schema', attr_nm='schema'
            )
            # fmt: on
        except ValueError as e:
            raise e
        restrictions = {
            **(restrictions or dict()),
            **{
                "lower(table_name)": f"'{table.lower()}'",
                "lower(table_schema)": f"'{schema.lower()}'",
            },
        }
        if all_schemas:
            _ = restrictions.pop("lower(table_schema)")

        sql = self._info_schema_generic(
            obj="column", fields=fields, restrictions=restrictions, order_by=order_by
        )

        return self._query(sql=sql) if self(run) else sql
    
    def columns(
        self,
        nm: Optional[str] = None,
        from_info_schema: bool = False,
        lower: bool = False,
        run: Optional[bool] = None,
    ) -> Union[str, List]:
        """Returns an ordered list of columns for a table or view.

        note:
            *   The default behavior of this method is to retrieve the columns
                for a table or view by selecting a single sample record
                from the table and extracting the column names directly off
                the returned :class:`pandas.DataFrame` due to the performance
                gains in selecting a sample record as opposed to querying the
                ``information_schema.columns``.
            *   This can be changed by passing `from_info_schema=False`.

        Args:
            nm (str):
                Name of table or view, including schema if the table or view is
                outside of the current schema.
            from_info_schema (bool):
                Indicates whether to retrieve columns via the
                ``information_schema.columns`` or by selecting a sample record
                from the table or view; defaults to `False`.
            lower (bool):
                Lower case each column in the list that's returned.
            run (bool):
                Indicates whether to execute ge sql or return as string;
                default is `True`.

        Returns (Union[str, List]):
            Either:
                1.  An ordered list of columns for the table or view, **or**
                2.  The query against the table or view as a :class:`str` of sql.

        """
        if from_info_schema:
            return self._columns_from_info_schema(nm=nm, lower=lower, run=run)
        else:
            return self._columns_from_sample(nm=nm, lower=lower, run=run)
        
    # -- Common Queries -------------------------------------------------------
    
    def select(
        self,
        nm: Optional[str] = None,
        n: Optional[int] = None,
        run: Optional[bool] = None,
    ) -> Union[str, pd.DataFrame]:
        """Select `n` sample records from a table.

        Args:
            nm (str):
                Name of table or view to sample, including schema if the table
                or view is outside of the current schema.
            n (int):
                Number of records to return, implemented as a 'limit' clause
                in the query; defaults to 1.
            run (bool):
                Indicates whether to execute ge sql or return as string;
                default is `True`.

        Returns (Union[str, pd.DataFrame]):
            Either:
                1.  The results of the query as a :class:`pandas.DataFrame`, or
                2.  The ge query as a :class:`str` of sql.

        """
        schema, nm = p(nm)
        try:
            # fmt: off
            schema, table = (
                self._validate(
                    val=(schema or self.schema), nm='schema', attr_nm='schema'
                ),
                self._validate(
                    val=(nm or self.nm), nm='nm', attr_nm='nm'
                )
            )
            # fmt: on
        except ValueError as e:
            raise e
        limit = f"limit {n or 1}" if n != -1 else str()
        sql = (
            f"""
            select
                *
            from {up(schema)}.{up(table)}
            {limit}
            """
        )
        return self._query(sql=sql) if self(run) else sql
     
    # noinspection PyBroadException
    def exists(self, nm: Optional[str] = None) -> bool:
        """Checks the existence of a table or view.

        Args:
            nm (str):
                Name of table or view, including schema if the table or view is
                outside of the current schema.

        Returns (bool):
            Boolean indication of whether or not the table or view exists.

        """
        try:
            _ = self.select(nm=nm, n=1)
            return True
        except:
            return False

    def is_distinct(self, nm: Optional[str] = None, field: Optional[str] = None) -> bool:
        """Checks if table `nm` is distinct on column `on_col`

        Args:
            nm (str):
                Table name.
            field (str):
                Column name.
                
        """
        try:
            return self.count(nm=nm, dst_of=field, as_perc=True) == 1
        except ValueError as e:
            raise e

    def count(
            self,
            nm: Optional[str] = None,
            of: Optional[str] = None,
            dst_of: Optional[str] = None,
            as_perc: Optional[bool] = None,
            run: Optional[bool] = None,
    ) -> Union[int, float]:
        """Number of records within a table or view.

        Args:
            nm (str):
                Table name, including schema if querying outside current schema.
            of (str):
                Column name (indistinct).
            dst_of (str):
                Column name (distinct).
            as_perc (bool):
                Option to return distinct count of the `dst_of` column as a
                percentage of the namespace depth of the table or view.
            run (bool):
                Indicates whether to execute ge sql or return as string;
                default is `True`.

        Returns (Union[str, pd.DataFrame]):
            Either:
                1.  The results of the query as a :class:`pandas.DataFrame`, or
                2.  The ge query as a :class:`str` of sql.

        """
        schema, nm = p(nm)
        try:
            obj_schema = self._validate(
                val=(schema or self.schema), nm='obj_schema', attr_nm='schema'
            )
            obj_name = self._validate(
                val=(nm or self.nm), nm='obj_name', attr_nm='obj_name'
            )
        except ValueError as e:
            raise e
        
        _from = f"from {obj_schema}.{obj_name}"
        sql = f"select count(*) {_from}"
        if of:
            sql = f"select count({of}) {_from}"
        if dst_of:
            sql = f"select count(distinct {dst_of}) {_from}"
        if as_perc:
            if not dst_of:
                raise ValueError(
                    "`as_perc=True` pr without specifying `dst_of` column."
                )
            sql = f"select count(distinct {dst_of}) / count(*) {_from}"
            
        return self._query(sql=sql, as_scalar=True) if self(run) else sql
    
    def show(
        self,
        obj: str,
        in_loc: Optional[str] = None,
        names: bool = False,
        run: Optional[bool] = None,
        **kwargs,
    ) -> Union[pd.DataFrame, List[str], str]:
        """Show schema objects of typ 'obj', optionally 'in_loc'.

        Args:
            obj (str):
                Schema object type ('tables', 'file formats', etc).
            in_loc (str):
                Snowflake location ('in schema sandbox', 'in database prod', etc).
            names (bool):
                Return a list of schema object names only ('name' field).
            run (bool):
                Execute the ge sql or return it as a string.

        Returns (Union[pd.DataFrame, str]):
            Either:
                1.  The results of the query as a :class:`pandas.DataFrame`
                2.  The 'names' column of the results returned as a list
                3.  The ge query as a :class:`str` of sql
            
        """
        _in_loc = f" in {in_loc}" if in_loc else str()
        sql = s(f"show {obj}{_in_loc}")
        if self(run):
            result = self._query(sql, **kwargs)
            return (
                result
                if not names
                else result.snf.to_list(
                    'name' if not kwargs.get('lower') else 'NAME'
                )
            )
        return self._query(sql, **kwargs) if self(run) else sql
    
    # -- Metadata -------------------------------------------------------------
    
    def ddl(
        self,
        nm: Optional[str] = None,
        obj: Optional[str] = None,
        run: Optional[bool] = None,
    ) -> str:
        """Query the DDL for an schema object.

        Args:
            nm (str):
                Name of the object to get DDL for, including schema if object
                is outside of the current schema.
            obj (str):
                Type of object to get DDL for (e.g. 'table', 'view', 'file-format').
            run (bool):
                Indicates whether to execute ge sql or return as string;
                default is `True`.

        Returns (str):
            Either:
                1.  The results of the query as a :class:`pandas.DataFrame`, or
                2.  The ge query as a :class:`str` of sql.

        """
        schema, nm = p(nm)
        try:
            # fmt: off
            obj, schema, nm = (
                self._validate(
                    val=(obj or self.obj), nm='obj', attr_nm='obj'
                ),
                self._validate(
                    val=(schema or self.schema), nm='schema', attr_nm='schema'
                ),
                self._validate(
                    val=(nm or self.nm), nm='nm', attr_nm='nm'
                )
            )
            # fmt: on
        except ValueError as e:
            raise e
        _nm = self._schema_object(nm=nm, schema=schema, typ=obj)
        sql = s(f"select get_ddl('{obj}', '{up(_nm)}') as ddl")
        return self._query(sql=sql).snf.to_list(n=1) if self(run) else sql
    
    def comment(
        self,
        nm: Optional[str] = None,
        obj: Optional[str] = None,
        set_as: Optional[str] = None,
        from_json: bool = False,
        as_json: bool = False,
        run: Optional[bool] = None,
        **kwargs,
    ) -> Union[str, Dict]:
        """Get or set comment on a schema object.

        Args:
            nm (str):
                Name of the schema object, including schema prefix if object
                is outside implicit scope of the current connection.
            obj (str):
                Type of schema object (e.g. 'table', 'schema', etc).
            set_as (str):
                Content to set as comment on schema object.
            from_json (bool):
                Parse schema object comment as a string of json and return it
                as a dictionary.
            as_json (bool):
                Dump contents of 'set_as' to a string of json prior to setting
                comment.
            run (bool):
                Indicates whether to execute generated sql or return as string;
                default is `True`.
            **kwargs:
                Keyword argument to pass to `json.loads(comment)` if
                *from_json=True*.

        Returns (Union[str, pd.DataFrame]):
            Either:
                1.  The schema object comment as a :class:`str`
                2.  The ge query as a :class:`str` of sql.
                3.  The schema object comment as a dictionary if *from_json=True*

        """
        # fmt: off
        if set_as:
            return self._set_comment(
                nm=nm,
                obj=obj,
                comment=set_as,
                as_json=as_json,
                run=run,
                **kwargs,
            )
        try:
            schema, nm = p(nm)
            obj, obj_schema, obj_name = (
                self._validate(
                    val=(obj or self.obj), nm='obj', attr_nm='obj'
                ).lower(),
                self._validate(
                    val=(schema or self.schema), nm='obj_schema', attr_nm='schema'
                ).lower(),
                self._validate(
                    val=(nm or self.nm), nm='obj_name', attr_nm='nm'
                ).lower()
            )
        except ValueError as e:
            raise e
        
        sql = self.info_schema(
            loc=self._cfg.sql.info_schema_loc(obj, stem=True),
            fields=['comment'],
            where=[
                f"lower({obj if obj != 'view' else 'table'}_name) = '{obj_name}'",
                self._c(
                    val=f"lower(table_schema) = '{obj_schema}'",
                    condition=bool(obj in ['table', 'view'])
                ),
            ],
            run=False,
        )
        
        if self(run):
            
            df = self._query(sql)
            
            if df.empty:
                raise SnowflakeObjectNotFound(
                    msg=(f"""
Object not found; \ndouble check that the schema object
exists and that the below sql is querying the intended
object:\n\n{sql}
"""
                         )
                )
            
            comment = df.snf.to_list('comment', n=1)
            return (
                comment if not from_json
                else
                json.loads(comment, **kwargs) if comment else dict()
            )
        
        # fmt: on
        return sql
    
    def last_altered(
        self, nm: Optional[str] = None, run: Optional[bool] = None
    ) -> Union[str, pd.Timestamp]:
        """Last altered timestamp for a table or view.

        Args:
            nm (str):
                Table name, including schema if creating a stage outside of the
                current schema.
            run (bool):
                Indicates whether to execute ge sql or return as string;
                default is `True`.

        Returns (Union[str, pd.DataFrame]):
            Either:
                1.  The results of the query as a :class:`pandas.DataFrame`, or
                2.  The ge query as a :class:`str` of sql.

        """
        try:
            sql = self.table_info(
                nm=nm,
                fields=["last_altered"],
                run=False,
            )
            return (
                self._query(sql=sql, as_scalar=True)
                if self(run)
                else sql
            )
        except AssertionError as e:
            raise e
    
    # -- Common DML Commands --------------------------------------------------

    def truncate(
        self, nm: Optional[str] = None, run: Optional[bool] = None
    ) -> Union[str, pd.DataFrame]:
        """Truncate a table.

        Args:
            nm (str):
                Name of table, including schema if the table is outside of the
                current schema.
            run (bool):
                Indicates whether to execute ge sql or return as string;
                default is `True`.

        Returns (Union[str, pd.DataFrame]):
            Either:
                1.  The results of the query as a :class:`pandas.DataFrame`, or
                2.  The ge query as a :class:`str` of sql.

        """
        schema, nm = p(nm)
        try:
            # fmt: off
            schema, name = (
                self._validate(
                    val=(schema or self.schema), nm='schema', attr_nm='schema'
                ),
                self._validate(
                    val=(nm or self.nm), nm='nm', attr_nm='nm'
                )
            )
            # fmt: on
        except ValueError as e:
            raise e
        sql = s(f"truncate table {up(schema)}.{up(name)}")
        return self._query(sql=sql) if self(run) else sql

    def drop(
        self,
        nm: Optional[str] = None,
        obj: Optional[str] = None,
        run: Optional[bool] = None,
    ) -> Union[str, pd.DataFrame]:
        """Drop a ``Snowflake`` object.

        Args:
            nm (str):
                Name of the object to drop, including schema if creating a stage
                outside of the current schema.
            obj (str):
                Type of object to drop (e.g. 'table', 'schema', etc)
            run (bool):
                Indicates whether to execute ge sql or return as string;
                default is `True`.

        Returns (Union[str, pd.DataFrame]):
            Either:
                1.  The results of the query as a :class:`pandas.DataFrame`, or
                2.  The ge query as a :class:`str` of sql.

        """
        schema, nm = p(nm)
        try:
            # fmt: off
            obj_schema, obj_name, obj = (
                self._validate(
                    val=(schema or self.schema), nm='obj_schema', attr_nm='schema'
                ),
                self._validate(
                    val=(nm or self.nm), nm='obj_name', attr_nm='obj_name'
                ),
                self._validate(
                    val=(obj or self.obj), nm='obj', attr_nm='obj'
                )
            )
            # fmt: on
        except ValueError as e:
            raise e
        _name = self._schema_object(nm=obj_name, schema=obj_schema, typ=obj)
        sql = s(f"drop {obj} if exists {up(_name)}")
        return self._query(sql=sql) if self(run) else sql
    
    def clone(
        self,
        nm: Optional[str] = None,
        to: Optional[str] = None,
        obj: Optional[str] = None,
        run: Optional[bool] = None,
        replace: bool = False,
    ) -> Union[str, pd.DataFrame]:
        """Clone a ``Snowflake`` object.

        Warnings:
            *   Make sure to read `Snowflake's documentation
                <https://docs.snowflake.com/en/sql-reference/sql/create-clone.html>`_
                for restrictions and considerations when cloning objects.

        Note:
            *   In this specific method, the value pr to ``nm`` and ``to``
                can be a single object name, a single schema, or both in the
                form of `obj_schema.obj_name` depending on the desired outcome.
            *   Additionally, **at least one of the** ``nm`` **or** ``to``
                **arguments must be pr**.
            *   The defaults for the target object are constructed such that
                users can **either**:
                    1.  Clone objects to *other* schemas that inherit the
                        source object's *name* without specifying so in the
                        ``to`` argument, **or**
                    2.  Clone objects within the *current* schema that inherit
                        the source object's *schema* without specifying so in
                        the ``to`` argument.
            *   If providing a schema without a name to either argument, prefix
                the value pr with `__` to signify it's a schema and not
                a lower-level object to be cloned.
                    *   e.g. providing `nm='sample_table'` and
                        `to='__sandbox'` will clone `sample_table` from the
                        current schema to `sandbox.sample_table`.
            *   An assertion error will be raised raised if neither argument
                is specified as *this would result in a command to clone an
                object and store it in an object that has the same name &
                schema as the object being cloned*.

        Args:
            nm (str):
                Name of the object to clone, including schema if cloning an
                object outside of the current schema.
            to (str):
                Target name for cloned object, including schema if cloning an
                object outside of the current schema.
            obj (str):
                Type of object to clone (e.g. 'table', 'view', 'file-format');
                defaults to `table`.
            run (bool):
                Indicates whether to execute ge sql or return as string;
                default is `True`.
            replace (bool):
                Indicates whether to replace an existing stage if pre-existing;
                default is `False`.

        Returns (Union[str, pd.DataFrame]):
            Either:
                1.  The results of the query as a :class:`pandas.DataFrame`, or
                2.  The ge query as a :class:`str` of sql.

        """
        try:
            # fmt: off
            schema, nm = p(nm)
            to_schema, to = p(nm=to)
            obj, schema, nm = (
                self._validate(
                    val=(obj or self.obj), nm='obj', attr_nm='obj'
                ),
                self._validate(
                    val=(schema or self.schema), nm='schema', attr_nm='schema'
                ),
                self._validate(
                    val=(nm or self.nm), nm='nm', attr_nm='nm'
                )
            )

            to_schema, to = (
                to_schema or schema,
                to or nm
            )
            if not to_schema and not to:
                raise ValueError(
                    "At least one of '__schema' or 'name` must be pr "
                    "in the 'to' argument of sql.clone()."
                )
            if nm == to and schema == to_schema:
                raise ValueError(
                    f"Target object name & schema mirrors source object name/schema. "
                    f"Please provide a different value `to`"
                )
            # fmt: on
        except ValueError as e:
            raise e
        _create = self._create(replace=replace)
        _src = self._schema_object(nm=nm, schema=schema, typ=obj)
        _target = self._schema_object(nm=to, schema=to_schema, typ=obj)
        sql = s(f"{_create} {obj} {up(_src)} clone {up(_target)}")
        return self._query(sql=sql) if self(run) else sql
    
    def _set_comment(
        self,
        comment: str,
        nm: Optional[str] = None,
        obj: Optional[str] = None,
        as_json: bool = False,
        run: Optional[bool] = None,
        **kwargs,
    ) -> str:
        """Drop a ``Snowflake`` object.

        Args:
            comment (str):
                Value to set as comment on the schema object.
            nm (str):
                Name of the schema object to set comment on, including schema
                prefix if it's a table, view, or file format located outside
                of the current schema.
            obj (str):
                Type of schema object (e.g. 'table', 'schema', etc)
            as_json (bool):
                Dump comment to a string of json before setting it on the schema
                object.
            run (bool):
                Indicates whether to execute ge sql or return as string;
                default is `True`.
            **kwargs:
                Keyword argument to pass to `json.dumps(comment, **kwargs)`
                if *as_json = True*.

        Returns (str):
            Db response from the statement as a string.

        """
        # fmt: off
        schema, nm = p(nm)
        try:
            obj, obj_schema, obj_name = (
                self._validate(
                    val=(obj or self.obj), nm='obj', attr_nm='obj'
                ).lower(),
                self._validate(
                    val=(schema or self.schema), nm='obj_schema', attr_nm='schema'
                ).lower(),
                self._validate(
                    val=(nm or self.nm), nm='obj_name', attr_nm='nm'
                ).lower()
            )
        except ValueError as e:
            raise e
        
        _comment = (
            comment
            if not as_json
            else json.dumps(comment, **kwargs)
        )
        _name = self._schema_object(nm=obj_name, schema=obj_schema, typ=obj)
        _sql = (
            f"""
            comment on {obj} {_name}
            is '{_comment}'
            """
        )
        sql = s(_sql, trailing=False, blanks=True)
        return self._query(sql, as_scalar=True) if self(run) else sql
    
    # -- Staging Operations ---------------------------------------------------

    def create_stage(
        self,
        nm_stage: str,
        nm_format: str,
        replace: bool = False,
        run: Optional[bool] = None,
    ) -> Union[str, pd.DataFrame]:
        """Create a staging table.

        Args:
            nm_stage (str):
                Name of stage to create, including schema if creating a stage
                outside of the current schema.
            nm_format (str):
                Name of file format to specify for the stage, including schema
                if using a format from outside of the current schema.
            run (bool):
                Indicates whether to execute ge sql or return as string;
                default is `True`.
            replace (bool):
                Indicates whether to replace an existing stage if pre-existing;
                default is `False`.

        Returns (Union[str, pd.DataFrame]):
            Either:
                1.  The results of the query as a :class:`pandas.DataFrame`, or
                2.  The ge query as a :class:`str` of sql.

        """
        create = self._create(replace=replace)
        sql = s(f"{create} stage {nm_stage} file_format = {nm_format};")
        return self._query(sql=sql) if self(run) else sql

    def put_file_from_stage(
        self,
        path: Union[Path, str],
        nm_stage: str,
        options: Optional[Dict] = None,
        ignore_defaults: bool = False,
        run: Optional[bool] = None,
    ) -> Union[str, pd.DataFrame]:
        """Generates a 'put' command into a staging table from a local file.

        Args:
            path (Union[Path, str]):
                Path to local data file as a :class:`pathlib.Path` or string.
            nm_stage (str):
                Name of the staging table to load into.
            run (bool):
                Indicates whether to execute ge sql or return as string;
                default is `True`.
            options (dict):
                Optional arguments to add to `put` statement in addition to
                the values specified in the ``loading.put`` section
                of **snowmobile.toml**.
            ignore_defaults (bool):
                Option to ignore the values specified in **snowmobile.toml**;
                defaults to `False`.

        Returns (Union[str, pd.DataFrame]):
            Either:
                1.  The results of the query as a :class:`pandas.DataFrame`, or
                2.  The ge query as a :class:`str` of sql.

        """
        path = Path(str(path))
        statement = [f"put file://{path.as_posix()} @{nm_stage}"]
        # fmt: off
        defaults = (
            dict() if ignore_defaults
            else self._cfg.loading.put.dict(by_alias=False)
        )
        options = {
            **defaults,
            **(options or dict()),
        }
        for k, v in options.items():
            statement.append(f"\t{k} = {str(v).lower() if isinstance(v, bool) else v}")
        # fmt: on
        _sql = "\n".join(statement)
        sql = s(_sql, trailing=False, whitespace=False, blanks=True)

        return self._query(sql=sql) if self(run) else sql

    def copy_into_table_from_stage(
        self,
        nm: str,
        nm_stage: str,
        options: Optional[Dict] = None,
        ignore_defaults: bool = False,
        run: Optional[bool] = None,
    ) -> Union[str, pd.DataFrame]:
        """Generates a command to copy data into a table from a staging table.

        Args:
            nm (str):
                Name of the object to drop, including schema if creating a stage
                outside of the current schema.
            nm_stage (str):
                Name of the staging table to load from.
            run (bool):
                Indicates whether to execute ge sql or return as string;
                default is `True`.
            options (dict):
                Optional arguments to add to `put` statement in addition to
                the values specified in the ``loading.put`` section
                of **snowmobile.toml**.
            ignore_defaults (bool):
                Option to ignore the values specified in **snowmobile.toml**;
                defaults to `False`.

        Returns (Union[str, pd.DataFrame]):
            Either:
                1.  The results of the query as a :class:`pandas.DataFrame`, or
                2.  The ge query as a :class:`str` of sql.

        """
        statement = [f"copy into {nm} from @{nm_stage}"]
        defaults = (
            self._cfg.loading.copy_into.dict(by_alias=False)
            if not ignore_defaults
            else dict()
        )
        options = {**defaults, **(options or dict())}
        for k, v in options.items():
            statement.append(f"\t{k} = {v}")
        _sql = "\n".join(statement)
        sql = s(_sql, trailing=False, whitespace=False, blanks=True)
        return self._query(sql=sql) if self(run) else sql

    # -- Current Connection Information ---------------------------------------

    def current(
        self, obj: str, run: Optional[bool] = None
    ) -> Union[str, Union[str, int]]:
        """Generic implementation of 'select current' for session-based objects.

        Args:
            obj (str):
                Type of object to retrieve information for (schema, session, ..).
            run (bool):
                Indicates whether to execute ge sql or return as string;
                default is `True`.

        Returns (Union[str, pd.DataFrame]):
            Either:
                1.  The results of the query as a :class:`pandas.DataFrame`, or
                2.  The ge query as a :class:`str` of sql.

        """
        _sql = f"select current_{obj}()"
        sql = s(_sql)
        return self._query(sql=sql).snf.to_list(n=1) if self(run) else sql

    def current_session(self, run: Optional[bool] = None) -> Union[str, pd.DataFrame]:
        """Select the current session."""
        return self.current(obj="session", run=run)

    def current_schema(self, run: Optional[bool] = None) -> Union[str, pd.DataFrame]:
        """Select the current schema."""
        return self.current(obj="schema", run=run)

    def current_database(self, run: Optional[bool] = None) -> Union[str, pd.DataFrame]:
        """Select the current database."""
        return self.current(obj="database", run=run)

    def current_warehouse(self, run: Optional[bool] = None) -> Union[str, pd.DataFrame]:
        """Select the current warehouse."""
        return self.current(obj="warehouse", run=run)

    def current_role(self, run: Optional[bool] = None) -> Union[str, pd.DataFrame]:
        """Select the current role."""
        return self.current(obj="role", run=run)

    # -- Generic 'use ___' Statements -----------------------------------------

    def use(self, nm: str, obj: str, run: Optional[bool] = None):
        """Generic implementation of 'use' command for schema objects.

        Args:
            nm (str):
                Name of object to use (schema name, warehouse name, role name, ..).
            obj (str):
                Type of object to use (schema, warehouse, role, ..).
            run (bool):
                Indicates whether to execute ge sql or return as string;
                default is `True`.

        Returns (Union[str, pd.DataFrame]):
            Either:
                1.  The results of the query as a :class:`pandas.DataFrame`, or
                2.  The ge query as a :class:`str` of sql.

        """
        # fmt: off
        try:
            name = self._validate(
                val=(nm or self.nm), nm='nm', attr_nm='nm'
            )
        except ValueError as e:
            raise e
        # fmt: on
        _sql = f"use {obj} {up(name)}"
        sql = s(_sql)
        return self._query(sql=sql) if self(run) else sql

    def use_schema(
        self, nm: Optional[str] = None, run: Optional[bool] = None
    ) -> Union[str, pd.DataFrame]:
        """Use schema command."""
        return self.use(obj="schema", nm=nm, run=run)

    def use_database(
        self, nm: Optional[str] = None, run: Optional[bool] = None
    ) -> Union[str, pd.DataFrame]:
        """Use database command."""
        return self.use(obj="database", nm=nm, run=run)

    def use_warehouse(
        self, nm: Optional[str] = None, run: Optional[bool] = None
    ) -> Union[str, pd.DataFrame]:
        """Use warehouse command."""
        return self.use(obj="warehouse", nm=nm, run=run)

    def use_role(
        self, nm: Optional[str] = None, run: Optional[bool] = None
    ) -> Union[str, pd.DataFrame]:
        """Use role command."""
        return self.use(obj="role", nm=nm, run=run)

    # -- Private and Static Methods -------------------------------------------

    @staticmethod
    def _validate(
        val: Optional[str, int], nm: str, attr_nm: Optional[str] = None
    ) -> str:
        """Validates the value of an argument passed to a method.

        This method is built to validate method arguments in instances where an
        unspecified argument can fall back to an attribute if it has been set.

        Each of the 'closing' variables below represents a different ending to
        a sentence within the exception message depending on the value pr
        from the method and if the attribute the argument falls back to has been
        set at the time the method is called.

        Args:
            val (Union[str, int]:
                Value to validate.
            nm (str):
                Name of argument in the method being called.
            attr_nm (str):
                Name of attribute to fall back to if the boolean representation
                of ``val`` is `False`.

        """
        if not val:
            closing1 = (
                "." if not attr_nm else f", nor is its fallback attribute '{attr_nm}'."
            )
            closing2 = (
                "."
                if not attr_nm
                else f" or set the '{attr_nm}' attribute before calling the method."
            )
            raise ValueError(
                f"\nValue pr for '{nm}' is not valid{closing1}\n"
                f"Please provide a valid value for '{nm}'{closing2}"
            )
        return val

    def _columns_from_info_schema(
        self, nm: Optional[str] = None, lower: bool = False, run: Optional[bool] = None
    ) -> Union[str, List]:
        """Retrieves list of columns for a table or view **from information schema**.

        Args:
            nm (str):
                Name of table or view, including schema if the table or view is
                outside of the current schema.
            lower (bool):
                Lower case each column in the list that's returned.
            run (bool):
                Indicates whether to execute ge sql or return as string;
                default is `True`.

        Returns (Union[str, List]):
            Either:
                1.  An ordered list of columns for the table or view, **or**
                2.  The query against ``information_schema.columns`` as a
                    :class:`str` of sql.

        """
        sql = self.column_info(
            nm=nm, fields=["ordinal_position", "column_name"], order_by=[1], run=False
        )
        if self(run):
            return [
                c.lower() if lower else c
                for c in self._query(sql).snf.to_list(col="column_name")
            ]
        else:
            return sql

    def _columns_from_sample(
        self, nm: Optional[str] = None, lower: bool = False, run: Optional[bool] = None
    ) -> Union[str, List]:
        """Retrieves a list of columns for a table or view from **sampling table**.

        Args:
            nm (str):
                Name of table or view, including schema if the table or view is
                outside of the current schema.
            lower (bool):
                Lower case each column in the list that's returned.
            run (bool):
                Indicates whether to execute ge sql or return as string;
                default is `True`.

        Returns (Union[str, List]):
            Either:
                1.  An ordered list of columns for the table or view, **or**
                2.  The query against the table or view as a :class:`str` of sql.

        """
        _sql = self.select(nm=nm, run=False, n=1)
        sql = s(_sql)
        if self(run):
            return [c for c in self._query(sql, lower=lower).columns]
        else:
            return sql

    def _info_schema_generic(
        self,
        obj: str,
        fields: List[str] = None,
        restrictions: Dict[str, str] = None,
        order_by: Optional[List] = None,
    ) -> str:
        """Generic case of selecting from information schema tables/columns.

        Queries different parts of the information schema based on an ``obj``
        and the mapping of object type to information schema defined in
        `snowmobile.core.sql._map_information_schema.py`.

        """
        info_schema_loc = self._cfg.sql.info_schema_loc(obj=obj)
        fields = self.fields(fields=fields)
        where = self.where(restrictions=restrictions)
        order_by = self.order(by=order_by)

        sql = f"""
select
{fields}
from {info_schema_loc}
{where}
{order_by}
"""
        return s(sql, trailing=False, blanks=True)

    @staticmethod
    def order(by: List[Union[int, str]]) -> str:
        """Generates 'order by' clause from a list of fields or field ordinal positions."""
        if by:
            order_by_fields = ",".join(str(v) for v in by)
            return f"order by {order_by_fields}"
        else:
            return str()

    @staticmethod
    def where(restrictions: Dict) -> str:
        """Generates a 'where' clause based on a dictionary of restrictions.

        Args:
            restrictions (dict):
                A dictionary of conditionals where each key/value pair
                respectively represents the left/right side of a condition
                within a 'where' clause.

        Returns (str):
            Formatted where clause.

        """
        if restrictions:
            args = [
                f"{str(where_this)} = {str(equals_this)}"
                for where_this, equals_this in restrictions.items()
            ]
            args = "\n\tand ".join(args)
            return f"where\n\t{args}"
        else:
            return str()

    @staticmethod
    def fields(fields: Optional[List[str]] = None) -> str:
        """Utility to generate fields within a 'select' statement."""
        return "\n".join(
            f'\t{"," if i > 1 else ""}{f}'
            for i, f in enumerate(fields or ["*"], start=1)
        )
    
    @staticmethod
    def _c(val: str, condition: bool) -> str:
        """Checks val based on a condition."""
        return val if condition else str()
    
    @staticmethod
    def _create(replace: bool = False):
        """Utility to generate 'create'/'create or replace' based on an argument."""
        return "create" if not replace else "create or replace"

    def _r(self, run: Union[bool, None]) -> bool:
        """Determines whether or not to execute a piece of sql.

        Used in all subsequent methods containing a `run` argument.

        Args:
            run (Union[bool, None]):
                The value from a method's `run` argument.

        Returns (bool):
            Boolean value indicating whether or not to execute the sql ge by
            the method to which the value of `run` was passed.

        note:
            *   The default value of `run` in all subsequent methods is ``None``.
            *   When any method of :class:`SQL` containing a `run` argument is called,
                the argument's value is passed to this method which returns either:
                    1.  The argument's value if it is a valid bool (i.e. a user-pr
                        value to the method), or
                    2.  The boolean representation of the current :attr:`auto_run`
                        attribute (`True` by default).

        """
        if isinstance(run, bool):
            return run
        else:
            return bool(self.auto_run)
    
    @staticmethod
    def _schema_object(nm: str, schema: str, typ: str) -> str:
        """Returns schema object as a string based on its 'typ'."""
        return (
            nm
            if typ.lower() not in ['table', 'view', 'file_format']
            else f"{schema}.{nm}"
        )

    def _reset(self) -> SQL:
        self.schema = self._cfg.connection.current.schema_name
        self.nm = None
        self.obj = "table"
        return self

    def __call__(self, run: bool) -> bool:
        return self._r(run)

    def __str__(self) -> str:
        return f"snowmobile.SQL(creds='{self._cfg.connection.creds}')"

    def __repr__(self) -> str:
        return f"snowmobile.SQL(creds='{self._cfg.connection.creds}')"



"""
Unit tests for :class:`ExceptionHandler`.
"""

import pytest

import time
from typing import Union

from snowmobile.core.errors import (
    StatementNotFoundError,
    DuplicateTagError,
    QAFailure,
    QAEmptyFailure,
    QADiffFailure,
    InternalError,
    Error,
)

errors = Union[
    StatementNotFoundError,
    DuplicateTagError,
    QAFailure,
    QAEmptyFailure,
    QADiffFailure,
    InternalError,
    Error,
]


@pytest.mark.exception_handler
def test_default_exception_handler_defaults(default_test_exception_handler):
    """Testing default values of ExceptionHandler object."""
    assert default_test_exception_handler.within is None
    assert isinstance(default_test_exception_handler.ctx_id, int)
    assert default_test_exception_handler.by_ctx
    assert not default_test_exception_handler.in_context
    assert default_test_exception_handler.outcome is None


@pytest.mark.exception_handler
def test_exception_handler_core(
    snowmobile_test_exception_handler,
    sn_delayed,
    any_unique_integer1,
    any_unique_integer2,
    any_exception_marked_not_to_be_raised1,
    any_exception_marked_not_to_be_raised2,
    to_be_raised_exception1,
    to_be_raised_exception2,
):
    """Kitchen sink testing of ExceptionHandler functionality.."""
    e = snowmobile_test_exception_handler

    any_dormant_exc1, typ_exc_d1 = any_exception_marked_not_to_be_raised1
    any_dormant_exc2, typ_exc_d2 = any_exception_marked_not_to_be_raised2
    any_raise_exc1, typ_exc_r1 = to_be_raised_exception1
    any_raise_exc2, typ_exc_r2 = to_be_raised_exception2

    assert e.within == type(sn_delayed)  # expected: snowmobile.Snowmobile

    def with_one_dormant_exception():
        """setting an initial `ctx_id`; collecting a dormant exception."""
        # -- setting e.ctx_id #1 --
        e.set(ctx_id=any_unique_integer1)
        assert e.ctx_id == any_unique_integer1

        # -- collect exception1 --
        e.collect(e=any_dormant_exc1)

        assert isinstance(e.last, typ_exc_d1)
        assert e.last == e.first

        assert e.seen(of_type=typ_exc_d1) == e.seen(of_type=[typ_exc_d1])

    def with_another_dormant_exception():
        """Collecting another dormant exception."""
        # -- collect exception2 --
        time.sleep(1)
        e.collect(e=any_dormant_exc2)

        assert e.get() == e.current == e.by_tmstmp

        assert isinstance(e.first, typ_exc_d1)
        assert isinstance(e.last, typ_exc_d2)
        assert e.last != e.first

        assert e.get(to_raise=True) == dict()
        assert not e.seen(to_raise=True)

    def with_an_additional_exception_to_be_raised():
        """Collecting an exception **to be raised**."""
        # -- collect exception3; FIRST TO BE RAISED --
        time.sleep(1)
        e.collect(e=any_raise_exc1)

        assert len(e.get()) == 3
        assert len(e.get(to_raise=True)) == 1
        assert e.get(to_raise=True, first=True) == e.get(to_raise=True, last=True)

        assert e.seen(to_raise=True, of_type=typ_exc_r1)
        assert not e.seen(to_raise=True, of_type=typ_exc_d1)

    def with_a_new_ctx_id_and_an_exception_to_be_raised():
        """Setting a new `ctx_id` and collecting an exception to **be raised**."""
        # -- set ctx_id --
        e.set(ctx_id=any_unique_integer2)

        assert e.ctx_id == any_unique_integer2

        assert e.seen(from_ctx=any_unique_integer1)
        assert not e.seen(from_ctx=any_unique_integer2)

        # -- collect exception4; first in current context --
        assert not e.seen(to_raise=True)
        e.collect(e=any_raise_exc2)
        assert e.seen(to_raise=True)

        assert e.get(to_raise=True, first=True) == e.get(to_raise=True, last=True)
        assert not e.seen(to_raise=False)

    def core_exception_querying_across_two_ctx_ids():
        """Test core behavior of querying exception information from two separate
        `ctx_id`(s)."""
        assert len(e.by_ctx) == 2

        assert len(e.get(from_ctx=any_unique_integer1)) == 3
        assert len(e.get(from_ctx=any_unique_integer1, to_raise=True)) == 1
        assert len(e.get(from_ctx=any_unique_integer1, to_raise=False)) == 2

        assert len(e.get(from_ctx=any_unique_integer2)) == 1
        assert len(e.get(from_ctx=any_unique_integer2, to_raise=True)) == 1
        assert len(e.get(from_ctx=any_unique_integer2, to_raise=False)) == 0

    with_one_dormant_exception()
    with_another_dormant_exception()
    with_an_additional_exception_to_be_raised()
    with_a_new_ctx_id_and_an_exception_to_be_raised()
    core_exception_querying_across_two_ctx_ids()



"""[sql] (**snowmobile-ext.toml**)"""
from __future__ import annotations

import re

from typing import Dict, List

from pydantic import Field

from .base import Base


class SQL(Base):
    """[sql] (**snowmobile-ext.toml**)"""

    # fmt: off
    generic_anchors: Dict = Field(
        default_factory=dict, alias="generic-anchors"
    )
    kw_exceptions: Dict = Field(
        default_factory=dict, alias="keyword-exceptions"
    )
    named_objects: List = Field(
        default_factory=list, alias="named-objects"
    )
    info_schema_exceptions: Dict[str, str] = Field(
        default_factory=dict, alias="information-schema-exceptions"
    )
    desc_is_simple: bool = Field(
        default=True, alias="desc-is-simple"
    )
    pr_over_ge: bool = Field(
        default=True, alias="pr-over-ge"
    )
    # fmt: on

    def info_schema_loc(self, obj: str, stem: bool = False) -> str:
        """Returns information schema table for object if other than making plural.

        i.e.:
            *   'tables'  -> 'tables'
            *   'table'  --> 'tables'
            *   'schemas' -> 'schemata'
            *   'schema' --> 'schemata'

        """
        obj = obj.rstrip("s")
        default = f"{obj}s"  # 'table' -> 'tables' / 'column' -> 'columns'
        _loc = f"information_schema.{self.info_schema_exceptions.get(obj, default)}"
        return _loc if not stem else _loc.split('.')[-1]

    def objects_within(self, first_line: str):
        """Searches the first line of sql for matches to named objects."""
        matched_terms = {
            i: re.findall(f"\\b{term}\\b", first_line)
            for i, term in enumerate(self.named_objects)
        }
        return {k: v[0] for k, v in matched_terms.items() if v}



"""
This file exists for the sole purpose of exporting different configurations of
markup_ddl.sql for testing purposes.

The relative 'markup_ddl.sql' path below assumes that this script is stored
within 'tests/func/data/sql/markup/.snowmobile' directory.
"""
from pathlib import Path
from typing import List, Dict, Tuple
import snowmobile


def get_testing_script(nm: str, results_limit: int = 1):
    path = Path(__file__).absolute().parent.parent / nm
    sn = snowmobile.connect(delay=True, config_file_nm="snowmobile_testing.toml")
    sn.cfg.script.markup.result_limit = results_limit
    return snowmobile.Script(path=path, sn=sn)


def export_markups(script_dtl: List[Tuple[str, bool, Dict]], args: List[Dict]):
    for s in script_dtl:
        script_nm, run, run_kwargs = s
        script = get_testing_script(nm=script_nm)
        if run:
            script.run(**run_kwargs)
        for arg in args:
            doc = script.doc()(**arg["config"])
            doc.save(**arg["save"])


args = [
    {"config": {"prefix": "(default) "}, "save": {}},
    {"config": {"prefix": "(no sql) "}, "save": {"md": True}},
    {
        "config": {"incl_markers": False, "prefix": "(no markers) "},
        "save": {},
    },
    {
        "config": {
            "incl_markers": False,
            "incl_exp_ctx": False,
            "prefix": "(no disclaimer, no markers) ",
        },
        "save": {"sql": True},
    },
]


script_dtl = [
    ("markup_no_results.sql", False, {}),
    ("markup_with_results.sql", True, {"on_failure": "c"}),
    ("markup_template_anchor.sql", False, {}),
]

export_markups(script_dtl=script_dtl, args=args)



"""
[connection] section from **snowmobile.toml**, including subsections.
"""
from __future__ import annotations

from typing import Dict

from pydantic import Field

from .base import Base


class Credentials(Base):
    """[connection.credentials.credentials_alias]"""

    # fmt: off
    _alias: str = Field(
        default_factory=str, alias="_alias"
    )
    user: str = Field(
        default_factory=str, alias='user',
    )

    password: str = Field(
        default_factory=str, alias='password',
    )

    role: str = Field(
        default_factory=str, alias='role',
    )

    account: str = Field(
        default_factory=str, alias='account',
    )

    warehouse: str = Field(
        default_factory=str, alias='warehouse',
    )

    database: str = Field(
        default_factory=str, alias='database',
    )

    schema_name: str = Field(
        default_factory=str, alias='schema',
    )
    # fmt: on

    def as_nm(self, n: str):
        """Sets the credentials alias."""
        self._alias = n
        return self

    @property
    def credentials(self):
        """Returns namespace as a dictionary, excluding :attr:`_alias`."""
        return {k: v for k, v in self.dict(by_alias=True).items() if k != "_alias"}

    def __str__(self):
        """Altering inherited str method to mask credentials detail."""
        return f"Credentials('{self._alias}')"

    def __repr__(self):
        return f"Credentials('{self._alias}')"


class Connection(Base):
    """[connection]

    This includes the :attr:`default_alias` which is the set of credentials
    that :mod:`snowmobile` will authenticate with if :attr:`creds` is not
    explicitly passed.

    Attributes:
        default_alias (str):
            The set of credentials that is used if :attr:`creds` is not
            explicitly passed to :class:`snowmobile.connect` on
            instantiation.
        creds (str):
            The name given to the set of credentials within the
            **credentials** block of the **snowmobile.toml** file (e.g.
            [credentials.creds] assigns an :attr:`creds` to a given set of
            credentials.
        creds (dict[str, Creds]):
            A dictionary of :attr:`creds` to the associated
            :class:`Creds` object containing its credentials.

    """

    # fmt: off
    default_alias: str = Field(
        default_factory=str, alias='default-creds'
    )
    provided_alias: str = Field(
        default_factory=str, alias='pr-creds'
    )
    credentials: Dict[str, Credentials] = Field(
        default_factory=dict, alias="stored-credentials"
    )
    defaults: Dict = Field(
        default_factory=dict, alias="default-arguments"
    )
    # fmt: on

    def __init__(self, **data):

        super().__init__(**data)

        for k, v in data["credentials"].items():
            self.credentials[k] = Credentials(**v).as_nm(n=k)

        if not self.default_alias:
            self.default_alias = list(self.credentials)[0]

    @property
    def creds(self):
        """Credentials alias used by current Connection."""
        return self.provided_alias or self.default_alias

    @property
    def current(self):
        """Returns current credentials."""
        try:
            return self.credentials[self.creds]
        except KeyError as e:
            raise e
        # return self.get(self.creds)

    @property
    def connect_kwargs(self) -> Dict:
        """Arguments from snowmobile.toml for `snowflake.connector.connect()`."""
        return {**self.defaults, **self.current.credentials}



"""
Demonstrate behavior of Connector's 'ensure_alive' parameter.
..docs/snippets/connector_ensure_alive.py
"""
import snowmobile

# --- SESSION #1 ---

# Explicitly providing default argument for clarity
sn = snowmobile.connect(ensure_alive=True)

print(sn.alive)  #> True
type(sn.con)     #> snowflake.connector.connection.SnowflakeConnection

# Storing 1st session ID
session1 = sn.current('session')

# Killing connection
sn.disconnect()

print(sn.alive)  #> False
type(sn.con)     #> NoneType

# --- SESSION #2 ---

# Calling any method requiring a connection
_ = sn.query("select 1")

# Storing 2nd session ID
session2 = sn.current('session')

# Verifying both session IDs are valid
print(type(session1))  #> str
print(type(session2))  #> str

# Verifying they're unique
print(session1 != session2)  #> True

# -- complete example; should run 'as is' --
# snowmobile-include



"""
Defines the Scope for a given statement; invoked internally as the
scope/context of statements being considered by a :class:`snowmobile.Script`
is being altered through
:meth:`script.filter()<snowmobile.core.script.Script.filter()>`
"""
from __future__ import annotations

import re
import time
from typing import Dict

from . import Generic


class Scope(Generic):
    """Handles the scope/context for :class:`Statement` objects and derived classes.

    Should never be interacted with from the user-facing API.

    Attributes:
        base (str):
            The left-most word within a statement wrap. For **generic**
            st this will be the `keyword` and for **QA** statements
            this will be the literal word ``qa``.
        component (str):
            The component within a given wrap that is being evaluated; this will
            be exactly **one** of `kw`, `obj`, `anchor`, `desc`, or `nm`.
        incl_arg (str):
            The keyword argument that would be used to exclude a given
            component;
                *   e.g. if :attr:`component` is `kw`, :attr:`incl_arg` would
                    be ``incl_kw``.
        excl_arg (str):
            The keyword argument that would be used to exclude a given
            component; this would be the same as the above example except
            the value would be ``excl_kw`` as opposed to ``incl_kw``.
        fallback_to (dict):
            The default values to fall back to for :attr:`incl_arg` and
            :attr:`excl_arg` if they are not passed as a keyword argument
            by the user in :class:`Script`; defaults to including the
            :attr:`base` and excluding an empty list.
        provided_args (dict):
            The set of keyword arguments pr at the time of the last call
            to :meth:`eval()`.
        check_against_args (dict):
            The set of keyword arguments checked against at the time of the
            last call to :meth:`eval()`; will use pr arguments if they
            exist and the arguments from :attr:`fallback_to` otherwise.
        is_included (bool):
            Name is included based on the results of the last call to
            :meth:`eval()`.
        is_excluded (bool):
            Name is excluded based on the results of the last call to
            :meth:`eval()`.

    """

    def __init__(self, arg: str, base: str):
        """Instantiates a :class:`Scope` object."""
        super().__init__()
        self.component: str = arg  # 'kw'
        self.base = base.lower()  # 'qa'
        self.incl_arg = f"incl_{arg}"  # 'incl_kw'
        self.excl_arg = f"excl_{arg}"  # 'excl_kw'
        self.fallback_to: Dict = {self.incl_arg: [base], self.excl_arg: list()}
        self.provided_args: Dict = dict()
        self.check_against_args: Dict = dict()
        self.is_included: bool = bool()
        self.is_excluded: bool = bool()
        self.evals: Dict[float, Dict] = dict()

    def parse_kwargs(self, **kwargs) -> None:
        """Parses all filter arguments looking for those that match its base.

        Looks for include/exclude arguments within kwargs, populating
        :attr:`provided_args` with those that were pr and populates
        :attr:`check_against_args` with the same values if they were pr
        and fills in defaults from :attr:`fallback_to` otherwise.

        Args:
            **kwargs:
                Keyword arguments passed to :class:`Script.filter()` (e.g.
                `incl_kw`, `excl_kw`, ..)
        """
        self.provided_args = {
            k: v for k, v in kwargs.items() if k in [self.incl_arg, self.excl_arg]
        }
        self.check_against_args = {
            k: (self.provided_args.get(k) or self.fallback_to[k])
            for k in [self.incl_arg, self.excl_arg]
        }

    def matches_patterns(self, arg: str) -> bool:
        """Returns indication of if :attr:`base` matches a given set of patterns.

        Args:
            arg (str):
                Will either be the value of :attr:`incl_arg` or
                :attr:`exclude_arg`.

        Returns (bool):
            Indication of whether any matches were found.

        """
        escaped = any(
            re.findall(pattern=re.escape(self.base.lower()), string=p.lower())
            for p in self.check_against_args[arg]
        )
        unescaped = any(
            re.findall(string=self.base.lower(), pattern=p.lower())
            for p in self.check_against_args[arg]
        )
        return any([escaped, unescaped])

    @property
    def included(self):
        """Name is included based on results of last :meth:`eval()`."""
        return self.is_included and not self.is_excluded

    def eval(self, **kwargs) -> bool:
        """Evaluates filter arguments and updates context accordingly.

        Updates the values of :attr:`is_included`, :attr:`is_excluded`, and
        :attr:`included`.

        Args:
            **kwargs:
                Keyword arguments passed to :class:`Script.filter()` (e.g.
                `incl_kw`, `excl_kw`, ..)

        Returns (bool):
            Indicator of whether or not the statement should be
            included/excluded based on the context/keyword arguments pr.

        """
        self.evals[time.time()] = kwargs
        self.parse_kwargs(**kwargs)
        self.is_included = self.matches_patterns(arg=self.incl_arg)
        self.is_excluded = self.matches_patterns(arg=self.excl_arg)
        return self.included

    def __str__(self) -> str:
        return (
            f"Scope(arg='{self.component}', base='{self.base}', "
            f"included={self.included})"
        )

    def __repr__(self) -> str:
        return f"Scope(arg='{self.component}', base='{self.base}')"

    def __bool__(self) -> bool:
        """Mirrors the state of :attr:`included`."""
        return self.included



"""
Tests for wrap parsing.
"""
import pytest

from test import script as get_script


# noinspection PyProtectedMember
def _setup_for_test_tag_from_stripped_line():
    """Gets test cases and generates IDs for statement tags."""
    # get script with wrap test cases
    script = get_script(script_name="no_tags.sql")

    # remove index from ge wrap
    for s in script._statements_all.values():
        s.incl_idx_in_desc = False

    # generate test cases
    test_cases = [
        (s.nm(ge=True), s.nm(pr=True)) for s in script.st.values()
    ]

    # generate IDs test cases/console output
    # ids = [
    #     f"FirstLine='{s.first_line}',Name='{s.nm}'" for s in script.st.values()
    # ]
    ids = [
        f"nm_ge='{t[1]}', nm_pr='{t[0]}'"
        for t in test_cases
    ]

    return ids, test_cases


ids, test_cases = _setup_for_test_tag_from_stripped_line()


@pytest.mark.names
@pytest.mark.parametrize("tags", test_cases, ids=ids)
def test_tag_from_stripped_line(sn, tags):
    """Testing wrap generation from sql st in no_tags.sql."""
    tag_generated, tag_expected = tags
    assert tag_generated == tag_expected


@pytest.fixture()
def a_sample_tag(sn):
    """Testing __setattr__ on Name.."""
    from snowmobile.core.name import Name

    # noinspection SqlResolve
    return Name(
        configuration=sn.cfg,
        nm_pr="test-wrap",
        sql="select * from sample_table",
        index=1,
    )


@pytest.mark.names
def test_set_item_on_tag(a_sample_tag):
    """Testing __setattr__ and __bool__ on Name.."""
    a_sample_tag.is_included = False
    a_sample_tag.__setitem__("is_included", False)
    assert not a_sample_tag.is_included
    assert not a_sample_tag


@pytest.mark.names
def test_repr_on_tag(a_sample_tag):
    """Testing __repr__ on Name.."""
    assert a_sample_tag.__repr__() == "statement.Name(nm='test-wrap')"



"""
Instantiate a delayed snowmobile.Snowmobile object and inspect configuration model.
../docs/snippets/configuration/inspect_configuration.py
"""
import snowmobile

sn = snowmobile.Snowmobile(delay=True)

type(sn.cfg)              #> snowmobile.core.configuration.Configuration
print(sn.cfg.location)    #  'path/to/your/snowmobile.toml'

type(sn.cfg.connection)   #> snowmobile.core.cfg.connection.Connection
type(sn.cfg.loading)      #> snowmobile.core.cfg.loading.Loading
type(sn.cfg.script)       #> snowmobile.core.cfg.script.Script
type(sn.cfg.sql)          #> snowmobile.core.cfg.other.SQL
type(sn.cfg.ext_sources)  #> snowmobile.core.cfg.other.Location

# -- complete example; should run 'as is' --
# snowmobile-include



"""
Calling :meth:`script.doc()<snowmobile.core.script.Script.doc()>` returns a
:class:`Markup` containing a :class:`~snowmobile.core.section.Section` for each
statement or marker within the script.

.. note::

    A :class:`Markup` instance, ``m``, returned by
    :meth:`script.doc()<snowmobile.core.script.Script.doc()>`, makes no
    modifications to the sql file read by ``script``

    Instead, ``m`` will generate and export the following two files:

    -   A sql file stripped of all untagged comments, limited to statements
        within the context of ``script`` at the time ``m`` was created
    -   A markdown representation of the code and markup associated with the
        same set of statements

    By default, these files are exported to a ``.snowmobile`` directory
    alongside the sql file that was read by the ``script``; the directory name
    to use for ge exports can be configured in
    :ref:`[script.export-dir-name]<script.export-dir-name>`

    If the target directory does not yet exist, it will be created as part of
    the export process invoked by :meth:`m.save()<Markup.save()>`


The :ref:`default markdown configuration<script.markdown>` yields a `.md` file
with the below structure::

        # Script Name.sql         [script name gets an 'h1' header]

        - **Key1**: *Value1*      [keys are bolded, values are italicized]
        - **Key2**: *Value2*      [same for all tags/attributes found]
        - ...

        **Description**           [except for the 'Description' section]
                                  [this is just a blank canvas of markdown..]
                                  [..but this is configurable]

        ## (1) create-table~dummy_name [contents get 'h2' level headers]

        - **Key1**: *Value1*      [identical formatting for st/markers]

        **Description**           [statement descriptions get one of these too]

        **SQL**                   [as does their rendered sql]
            ...sql
                ...
                ...
            ...


        ## (2) update-table~dummy_name2

        ...                       [structure repeats for all contents in the script]


"""
from __future__ import annotations

import os
from pathlib import Path
from typing import Dict, List, Optional, Tuple, Union

from . import Generic, Configuration
from .connection import Snowmobile
from . import Diff, Empty, Section, Statement
from .paths import DIR_PKG_DATA
from .cfg import Marker
from .utils import Console


class Markup(Generic):
    """Contains all sections within the context of a :class:`~snowmobile.Script`.

    Parameters:
        sn (Snowmobile):
            A :class:`~snowmobile.Snowmobile` instance.
        path (Path):
            A full path to the sql file that script was instantiated from.
        contents (Dict[int, Union[Statement, Marker]]):
            A dictionary of the script's contents (st and markers) by
            index position.
        nm (Optional[str]):
            Alternate file name to use; defaults to ``path.name``.
        prefix (Optional[str]):
            Prefix to prepend to original file name when exporting.
        suffix (Optional[str]):
            Suffix to append to original file name when exporting.
        root_dir (Optional[Union[str, Path]]):
            Alternate target directory for exports; defaults to
            ``./.snowmobile`` where ``.`` is the directory containing the sql
            file that the ``script`` was created from.
        sub_dir (Optional[str]):
            Alternate sub-directory name; defaults to ``path.name`` where
            ``path`` is a full :class:`~pathlib.Path` to the sql file that the
            ``script`` was created from.
        incl_sql (bool):
            Include statements in export.
        incl_markers (bool):
            Include markers in export.
        incl_sql_tag (bool):
            Include the raw wrap in the sql that is rendered in the `md` export.
        incl_exp_ctx (bool):
            Include (configurable) disclaimer at the top of exported `sql` file.

    Attributes:
        exported (List[Path]):
            List of file paths that current instance has exported to.
        created (List[Path]):
            List of directory paths that current instance has created (should
            mostly apply for initial scaffolding build on first run only).

    """
    def __init__(
        self,
        sn: Snowmobile,
        path: Path,
        contents: Dict[int, Union[Statement, Marker]],
        nm: Optional[str] = None,
        prefix: Optional[str] = None,
        suffix: Optional[str] = None,
        root_dir: Optional[Union[str, Path]] = None,
        sub_dir: Optional[str] = None,
        incl_sql: bool = True,
        incl_markers: bool = True,
        incl_sql_tag: bool = False,
        incl_exp_ctx: bool = True,
        result_wrap: Optional[str] = None,
    ):
        super().__init__()

        # Private Attributes
        self._stdout = Console()
        self._pkg_data_dir = DIR_PKG_DATA
        self._cfg: Configuration = sn.cfg

        # Core Attributes
        self.path = Path(str(path))
        self.contents: Dict[int, Union[Statement, Marker]] = contents

        # Include / Exclude Components
        self.incl_sql = incl_sql
        self.incl_exp_ctx: bool = incl_exp_ctx
        self.incl_markers: bool = incl_markers
        self.incl_sql_tag = incl_sql_tag

        # Optional File (Name) Modifications
        self.nm: str = nm or self.path.name
        self.prefix: str = prefix or str()
        self.suffix: str = suffix or str()

        # Optional Directory Modifications
        self.root_dir: Path = Path(str(root_dir)) if root_dir else self.path.parent
        self.sub_dir: str = sub_dir or self._cfg.script.export_dir_nm

        # Files `exported` / Directories `created`
        self.exported: List[Path] = list()
        self.created: List[Path] = list()
        
        # Query Result Formatting
        self.result_wrap = result_wrap or str()

    @property
    def export_dir(self) -> Path:
        """Documentation sub-directory; `.snowmobile` by default."""
        return self.root_dir / self.sub_dir

    @property
    def _file_nm_components(self) -> Tuple[str, str]:
        """Utility for easy access to the stem and the extension of file name."""
        stem, _, ext = self.nm.rpartition(".")
        return stem, ext

    @property
    def _file_nm_sql(self) -> str:
        """Adjusted file name of the exported sql script."""
        stem, ext = self._file_nm_components
        return f"{self.prefix}{stem}{self.suffix}.{ext}"

    @property
    def _file_nm_md(self) -> str:
        """Adjusted file name of the exported markdown."""
        stem, ext = self._file_nm_components
        return f"{self.prefix}{stem}{self.suffix}.md"

    @property
    def _script_dir(self) -> Path:
        """Directory for all exports from specific nm."""
        stem, _, _ = self.nm.rpartition(".")
        return self.export_dir / stem

    @property
    def _path_md(self) -> Path:
        """Full path to write markdown to."""
        return self._script_dir / self._file_nm_md

    @property
    def _path_sql(self) -> Path:
        """Full path to write sql """
        return self._script_dir / self._file_nm_sql

    @property
    def sections(self) -> Dict[int, Section]:
        """Dictionary of all :class:`sections<snowmobile.core.Section>` by index position.
        """
        sections = {}
        for i, s in self.contents.items():
            if self._is_statement(s=s):  # create section from statement.section()
                sections[i] = s.as_section(self.incl_sql_tag)
            else:
                sections[i] = Section(  # create section from marker metadata
                    incl_sql_tag=self.incl_sql_tag,
                    is_multiline=True,
                    cfg=self._cfg,
                    raw=s.raw,
                    **s.as_args(),
                )

        return {i: sections[i] for i in sorted(sections)}

    @property
    def markdown(self) -> str:
        """Full markdown file as a string."""
        included = self._included
        return "\n\n".join(s.md for i, s in self.sections.items() if i in included)

    @property
    def _export_disclaimer(self) -> str:
        """Block comment disclaimer of save at top of exported sql file."""
        path_to_sql_txt = self._pkg_data_dir / "sql_export_heading.txt"
        with open(path_to_sql_txt, "r") as r:
            header = r.read()
        return f"{header}" if self.incl_exp_ctx else str()

    @property
    def _included(self):
        """All included indices based on incl_ attributes."""
        return {
            i
            for i, s in self.contents.items()
            if (
                (self.incl_sql and self._is_statement(s=s))
                or (self.incl_markers and not self._is_statement(s=s))
            )
        }

    @property
    def sql(self):
        """SQL for save."""
        to_export = [
            s.trim()
            if self._is_statement(s)
            else self._cfg.script.as_parsable(raw=s.raw, is_marker=True)
            for i, s in self.contents.items()
            if i in self._included
        ]
        if self.incl_exp_ctx:
            to_export.insert(0, self._export_disclaimer)
        return "\n".join(to_export)

    @staticmethod
    def _is_statement(s: Union[Statement, Diff, Empty, Marker]) -> bool:
        """Utility to check if a given instance of contents is a statement."""
        return isinstance(s, (Statement, Diff, Empty))

    def _scaffolding(self) -> None:
        """Ensures directory scaffolding exists before attempting save."""
        if not self._script_dir.exists():
            self._script_dir.mkdir(parents=True)
            self.created.append(self._script_dir)

    def _export(self, path: Path, val: str):
        """Ensure directory scaffolding exists and writes a string to a path (.sql or .md)."""
        self._scaffolding()
        if path.exists():
            os.remove(path)
        with open(path, "w", encoding='utf-8') as f:
            f.write(val)
            self.exported.append(path)
            self._stdout.offset_path(
                file_path=path, root_dir_nm=path.parent.name, indent="\t", output=True
            )

    def save(self, md: bool = True, sql: bool = True) -> None:
        """Save files to disk.

        Args:
            md (bool): Export a ge markdown file.
            sql (bool): Export a ge sql file.

        """
        if md:
            self._export(path=self._path_md, val=self.markdown)
        if sql:
            self._export(path=self._path_sql, val=self.sql)

    def __call__(self, **kwargs):
        """Batch setattr function for all keywords matching Markup's attributes."""
        return self._cfg.batch_set_attrs(obj=self, attrs=kwargs)

    def __str__(self) -> str:
        return f"snowmobile.core.Markup('{self._file_nm_sql}')"

    def __repr__(self) -> str:
        return f"snowmobile.core.Markup('{self._file_nm_sql}')"



"""
Base class for all :class:`Statement` objects.
"""
from __future__ import annotations

import time
from typing import Any, Callable, Dict, Optional, Union, Tuple, List

import pandas as pd
import sqlparse

from pydantic import BaseModel, Field

from pandas.io.sql import DatabaseError as pdDataBaseError
from snowflake.connector.errors import DatabaseError, ProgrammingError

from . import ExceptionHandler, Section, Name, errors, cfg
from . import Generic  # isort: skip
from .tag import Attrs
from .connection import Snowmobile


class Time(BaseModel):
    """
    Container for execution time info.
    """
    
    #: int: Unix timestamp statement is started
    started: int = Field(default_factory=int)
    
    #: int: Unix timestamp statement is completed
    ended: int = Field(default_factory=int)
    
    def __init__(self, **data):
        super().__init__(**data)
    
    def __int__(self):
        """Execution time in seconds."""
        return int(self.ended - self.started)
    
    def __str__(self):
        """Execution time as a string for console output."""
        return (
            f"{int(self)}s"
            if int(self) < 60
            else f"{int(int(self) / 60)}m"
        )


class Statement(Attrs, Name, Generic):
    """Base class for all :class:`Statement` objects.

    Home for attributes and methods that are associated with **all** statement
    objects, generic or QA.

    Attributes:
        sn (snowmobile.connect):
            :class:`snowmobile.connect` object.
        statement (Union[sqlparse.sql.Statement, str]):
            A :class:`sqlparse.sql.Statement` object.
        index (int):
            The context-specific index position of a statement within a script;
            can be `None`.
        patterns (config.Pattern):
            :class:`config.Pattern` object for more succinct access to
            values specified in **snowmobile.toml**.
        results (pd.DataFrame):
            The results of the statement if executed as a :class:`pandas.DataFrame`.
        outcome (int):
            Numeric indicator of outcome; defaults to `0` and is modified
            based on the outcome of statement execution and/or QA validation
            for derived classes.
        outcome_txt (str):
            Plain text of outcome ('skipped', 'failed', 'completed', 'passed').
        outcome_html (str):
            HTML text for the outcome as an admonition/information banner
            based on the following mapping of :attr:`outcome_txt` to
            admonition argument:
                *   `failed` ------> `warning`
                *   `completed` --> `info`
                *   `passed` -----> `success`
        start_time (int):
            Unix timestamp of the query start time if executed; 0 otherwise.
        end_time (int):
            Unix timestamp of the query end time if executed; 0 otherwise.
        execution_time (int):
            Execution time of the query in seconds if executed; 0 otherwise.
        execution_time_txt (str):
            Plain text description of execution time if executed; returned in
            seconds if execution time is less than 60 seconds, minutes otherwise.
        first_keyword (sqlparse.sql.Token):
            The first keyword within the statement as a :class:`sqlparse.sql.Token`.
        sql (str):
            The sql associated with the statement as a raw string.

    """

    # fmt: off
    PROCESS_OUTCOMES: Dict[Any, Any] = {
        0: ("-", ""),
        
        # -- start: base
        1: ("-", "error: execution"),
        2: ("success", "completed"),
        # -- end: base
        
        # -- start: derived
        -3: ("success", "passed"),
        -2: ("warning", "failed"),
        -1: ("-", "error: post-processing"),
        # -- end: derived
        
    }

    DERIVED_FAILURE_MAPPING = {
        'qa-diff': errors.QADiffFailure,
        'qa-empty': errors.QAEmptyFailure
    }
    # fmt: on

    def __init__(
        self,
        sn: Snowmobile,
        statement: Union[sqlparse.sql.Statement, str],
        index: Optional[int] = None,
        attrs_raw: Optional[str] = None,
        e: Optional[ExceptionHandler] = None,
        **kwargs,
    ):
        Generic.__init__(self)

        self._index: int = index
        self._outcome: int = int()

        self.outcome: bool = True
        self.executed: bool = bool()

        # self.sn = sn
        self.statement: sqlparse.sql.Statement = sn.cfg.script.ensure_sqlparse(
            statement
        )

        self.patterns: cfg.Pattern = sn.cfg.script.patterns
        self.results: pd.DataFrame = pd.DataFrame()

        #: Time: Execution time info
        self.time: Time = Time()
        
        self.start_time: int = int()
        self.end_time: int = int()
        self.execution_time: int = int()
        self.execution_time_txt: str = str()

        Attrs.__init__(self, sn=sn, raw=attrs_raw)
        self._sql = sn.cfg.script.strip_comments(s=self.statement)
        parsed, self._nm_intl = self.parse()
        self.update(parsed)
        
        Name.__init__(
            self,
            index=index,
            sql=self.sql(),
            nm_pr=self._nm_intl,
            configuration=self.sn.cfg,
        )

        self.e = e or ExceptionHandler(within=self)

    def sql(
        self, set_as: Optional[str] = None, tag: bool = False,
    ) -> Union[str, Statement]:
        """Raw sql from statement, including result limit if enabled."""
        if set_as:
            self._sql = set_as
            return self
        if tag:
            _attrs = self.tag(raw=False, namespace=True)
            if _attrs:
                _tag = self.sn.cfg.script.tag_from_attrs(
                    attrs=_attrs,
                    nm=_attrs.get('name', self.nm()),
                    wrap=False,
                )
            else:
                _tag = self.nm()
            tag = self.sn.cfg.script.wrap(_tag)
            return f"{tag}\n{self._sql}"
        if (
            self.sn.cfg.script.result_limit in [-1, 0]
            or self._sql.split('\n')[-1].strip().startswith('limit')
            or not self._sql.split('\n')[0].strip().lower().startswith('select')
        ):
            return self._sql
        return f"{self._sql}\nlimit {self.sn.cfg.script.result_limit}"

    def parse(self) -> Tuple[Dict, str]:
        """Parses tag contents into a valid dictionary.

        Uses the values specified in **snowmobile.toml** to parse a
        raw string of statement attributes into a valid dictionary.

        note:
            *   If :attr:`is_multiline` is `True` and `name` is not included
                within the arguments, an assertion error will be thrown.
            *   If :attr:`is_multiline` is `False`, the raw string within
                the wrap will be treated as the name.
            *   The :attr:`wrap` attribute is set once parsing is completed
                and name has been validated.

        Returns (dict):
            Parsed wrap arguments as a dictionary.

        """
        if not self.is_tagged:
            return dict(), str()

        if self.is_multiline:
            attrs_parsed = self.sn.cfg.script.parse_str(block=self.tag(raw=True))
            if "name" in attrs_parsed:
                name = attrs_parsed.pop("name")
            else:
                try:
                    name = self.sn.cfg.script.parse_name(raw=self.tag(raw=True))
                except errors.InvalidTagsError as e:
                    raise e

            return attrs_parsed, name

        return dict(), self.tag(raw=True)

    def start(self):
        """Sets :attr:`start_time` attribute."""
        self.start_time = time.time()

    def end(self):
        """Updates execution time attributes.

        In namespace, sets:
            *   :attr:`end_time`
            *   :attr:`execution_time`
            *   :attr:`execution_time_txt`

        """
        self.time.ended = self.end_time = time.time()
        self._outcome, self.outcome, self.executed = 2, True, True
        self.execution_time = int(self.end_time - self.start_time)
        self.execution_time_txt = (
            f"{self.execution_time}s"
            if self.execution_time < 60
            else f"{int(self.execution_time/60)}m"
        )

    def trim(self) -> str:
        """Statement as a string including only the sql and a single-line wrap name.

        note:
            The wrap name used here will be the user-pr wrap from the
            original script or a ge :attr:`Name.nm` if a wrap was not
            pr for a given statement.

        """
        _open, _close = self.sn.cfg.script.tag()
        return f"{_open}{self.nm()}{_close}\n{self.sql()};\n"

    @property
    def is_derived(self):
        """Indicates whether or not it's a generic or derived (QA) statement."""
        return self.anchor() in self.sn.cfg.QA_ANCHORS

    @property
    def lines(self) -> List[str]:
        """Returns each line within the statement as a list."""
        return self.sql().split("\n")

    def as_section(self, incl_sql_tag: Optional[bool] = None, result_wrap: Optional[str] = None) -> Section:
        """Returns current statement as a :class:`Section` object."""
        attrs = self.tag(namespace=True)
        results = attrs.get('results')
        if results and results.empty:
            _ = attrs.pop('results')
            
        return Section(
            index=self.index,
            h_contents=self.nm(),
            parsed=self.tag(namespace=True),
            raw=self.tag(raw=True),
            sql=self.sql(),
            cfg=self.sn.cfg,
            results=self.results,
            incl_sql_tag=incl_sql_tag,
            is_multiline=self.is_multiline,
            result_wrap=result_wrap,
        )

    def set_state(
        self,
        index: Optional[int] = None,
        ctx_id: Optional[int] = None,
        in_context: Optional[bool] = None,
        filters: dict = None,
    ) -> Statement:
        """Sets current state/context on a statement object.

        Args:
            ctx_id (int):
                Unix timestamp the :meth:`script.filter()` context manager was
                invoked.
            filters (dict):
                Kwargs passed to :meth:`script.filter()`.
            index (int):
                Integer to set as the statement's index position.

        """
        if index:
            self.index = index
        if ctx_id:
            self.e.set(ctx_id=ctx_id)
        if isinstance(in_context, bool):
            self.e.set(in_context=in_context)
        if filters:
            super().scope(**filters)
        return self

    def reset(
        self,
        index: bool = False,
        ctx_id: bool = False,
        in_context: bool = False,
        scope: bool = False,
    ) -> Statement:
        """Resets attributes on the statement object to reflect as if read from source.

        In its current form, includes:
            *   Resetting the statement/wrap's index to their original values.
            *   Resetting the :attr:`is_included` attribute of the statement's
                :attr:`wrap` to `True`.
            *   Populating :attr:`error_last` with errors from current context.
            *   Caching current context's timestamp and resetting back to `None`.

        """
        if index:
            self.index = self._index
        if ctx_id:
            self.e.reset(ctx_id=True)
        if in_context:
            self.e.reset(in_context=True)
        if scope:
            super().scope(**{})
        return self

    def process(self):
        """Used by derived classes for post-processing the returned results."""
        return self

    def run(
        self,
        as_df: bool = True,
        lower: bool = True,
        render: bool = False,
        on_error: Optional[str] = None,
        on_exception: Optional[str] = None,
        on_failure: Optional[str] = None,
        ctx_id: Optional[int] = None,
    ) -> Statement:
        """Run method for all statement objects.

        Args:
            as_df (bool):
                Store results of query as :class:`pandas.DataFrame` or
                :class:`SnowflakeCursor`.
            lower (bool):
                Lower case column names in :attr:`results` DataFrame if
                `as_df=True`.
            render (bool):
                Render the sql executed as markdown.
            on_error (str):
                Behavior if an execution/database error is encountered
                    * `None`: default behavior, exception will be raised
                    * `c`: continue with execution
            on_exception (str):
                Behavior if an exception is raised in the **post-processing**
                of results from a derived class of :class:`Statement` (
                :class:`Empty` and :class:`Diff`).
                    * `None`: default behavior, exception will be raised
                    * `c`: continue with execution
            on_failure (str):
                Behavior if no error is encountered in execution or post-processing
                but the result of the post-processing has turned the statement's
                :attr:`outcome` attribute to False, indicating the results
                returned by the statement have failed validation.
                    * `None`: default behavior, exception will be raised
                    * `c`: continue with execution

        Returns (Statement):
            Statement object post-executing query.

        """

        self.e.set(ctx_id=(ctx_id or -1))
        try:
            if self:
                self.start()
                self.time.started = time.time()
                self.results = self.sn.query(self.sql(), as_df=as_df, lower=lower)
                self.end()
                self.e.set(outcome=2)

        except (ProgrammingError, pdDataBaseError, DatabaseError) as e:
            self.e.collect(e=e).set(outcome=1)

        finally:  # only when execution did not raise database error
            if self.e.outcome != 1:
                self.process()
                
        # fmt: off
        if (
            self.e.seen(             # db error raised during execution -------
                of_type=errors.db_errors, to_raise=True
            )
            and on_error != "c"      # stop on execution error ----------------
        ):
            raise self.e.get(
                of_type=errors.db_errors,
                to_raise=True,
                first=True,
            )
        
        if (
            self.e.seen(             # post-processing error occurred ---------
                of_type=errors.StatementPostProcessingError,
                to_raise=True,
            )
            and on_exception != "c"  # stop on post-processing exception ------
        ):
            raise self.e.get(
                of_type=errors.StatementPostProcessingError,
                to_raise=True,
                first=True,
            )
        
        if (
            self.is_derived        # is child class with `.process()` method --
            and not self.outcome   # outcome of `.process()` did not pass -----
            and on_failure != "c"  # stop on failure of `.process()` ----------
        ):
            raise self.e.get(
                of_type=list(self.DERIVED_FAILURE_MAPPING.values()),
                to_raise=True,
                first=True,
            )
        # fmt: on

        if render:
            self.render()

        return self

    @staticmethod
    def _validate_parsed(attrs_parsed: Dict):
        """Returns args to verify 'name' attribute is present in a multiline wrap."""
        condition, msg = (
            attrs_parsed.get("name"),
            f"Required attribute 'name' not found in multi-line wrap's "
            f"arguments;\n attributes found are: {','.join(list(attrs_parsed))}",
        )
        return condition, msg

    def outcome_txt(self, _id: Optional[int] = None) -> str:
        """Outcome as a string."""
        return self.PROCESS_OUTCOMES[_id or self.e.outcome or 0][1]

    @property
    def outcome_html(self) -> str:
        """Outcome as an html admonition banner."""
        # TODO: Move this to patterns
        alert = self.PROCESS_OUTCOMES[self.e.outcome or 0][0]
        return f"""
<div class="alert-{alert}">
<center><b>====/ {self.outcome_txt()} /====</b></center>
</div>""".strip()

    def __len__(self):
        """Number of lines in the statement."""
        return len(self.lines)

    def __bool__(self):
        """Determined by the value of :attr:`Name.is_included`."""
        return self.is_included

    def __str__(self) -> str:
        return f"Statement('{self.nm()}')"

    def __repr__(self) -> str:
        return f"Statement('{self.nm()}')"



"""
:class:`~snowmobile.core.connection.Snowmobile` is the core of
:xref:`snowmobile`'s object model and a given instance is often shared across
multiple objects at once.

It is the primary method of executing statement against the warehouse and
it stores the fully parsed & validated ``snowmobile.toml`` file it was
instantiated with as its :attr:`~snowmobile.core.connection.Snowmobile.cfg`
attribute.

Within ``snowmobile``'s code and documentation, it is referred to as ``sn``
for brevity.

"""
from __future__ import annotations

from pathlib import Path
from typing import Optional, Union

import pandas as pd
from pandas.io.sql import DatabaseError as pdDataBaseError
from snowflake.connector import connect
from snowflake.connector.connection import SnowflakeCursor
from snowflake.connector import (
    DictCursor,
    SnowflakeConnection,
    DatabaseError,
    ProgrammingError,
)

from snowmobile.core import sql
from snowmobile.core import ExceptionHandler
from snowmobile.core.snowframe import SnowFrame

from . import Configuration


class Snowmobile(sql.SQL):
    """Primary method of statement execution and accessor to parsed snowmobile.toml.

    Parameters:

        creds (Optional[str]):
            Alias for the set of credentials to authenticate with; default
            behavior will fall back to the ``connection.default-creds``
            specified in `snowmobile.toml`, `or the first set of credentials
            stored if this configuration option is left blank`.
        delay (bool):
            Optionally delay establishing a connection when the object is
            instantiated, enabling access to the configuration object model
            through the :attr:`Connection.cfg` attribute; defaults to `False`.
        ensure_alive (bool):
            Establish a new connection if a method requiring a connection
            against the database is called while :attr:`alive` is `False`;
            defaults to `True`.
        config_file_nm (Optional[str]):
            Name of configuration file to use; defaults to `snowmobile.toml`.
        from_config (Optional[str, Path]):
            A full path to a specific configuration file to use; bypasses any
            checks for a cached file location and can be useful for container-based
            processes with restricted access to the local file system.
        **connect_kwargs:
            Additional arguments to provide to :xref:`snowflake.connector.connect()`;
            arguments pr here will over-ride connection arguments specified
            in `snowmobile.toml`, including:
                *   Connection parameters in `connection.default-arguments`
                *   Credentials parameters associated with a given alias
                *   Connection parameters associated with a given alias


    Attributes:

        cfg (snowmobile.core.configuration.Configuration):
            :class:`snowmobile.Configuration` object, which represents a fully
            parsed/validated `snowmobile.toml` file.
        con (SnowflakeConnection):
            :xref:`SnowflakeConnection` object; this attribute is populated
            when a connection is established and can be `None` if the
            :class:`Snowmobile` object was instantiated with `delay=True`.
        ensure_alive (bool):
            Establish a new connection if a method requiring a connection
            against the database is called while :attr:`alive` is `False`;
            defaults to `True`.
        e (ExceptionHandler):
            A :class:`ExceptionHandler<snowmobile.core.exception_handler.ExceptionHandler`
            class for orchestrating exceptions across objects; kept as a
            public attribute on the class as examining its contents can be
            helpful in debugging database errors.

    """

    def __init__(
        self,
        creds: Optional[str] = None,
        delay: bool = False,
        ensure_alive: bool = True,
        config_file_nm: Optional[str] = None,
        from_config: Optional[str, Path] = None,
        silence: bool = False,
        **connect_kwargs,
    ):

        # Parsed snowmobile.toml
        self.cfg: Configuration = Configuration(
            creds=creds,
            config_file_nm=config_file_nm,
            from_config=from_config,
            silence=silence,
        )

        # snowmobile.core.sql.SQL; inherits SQL and Generic
        super().__init__(_query_func=self.query, _cfg=self.cfg)
        
        # Snowflake Attributes; con = `None` until set by Snowmobile.connect()
        self.con: Optional[SnowflakeConnection] = None

        # Exception / Context Management
        self.e: ExceptionHandler = ExceptionHandler(within=self).set(ctx_id=-1)

        # Connection
        self.ensure_alive = ensure_alive

        if not delay:
            self.connect(**connect_kwargs)

    def connect(self, **kwargs) -> Snowmobile:
        """Establishes connection to Snowflake.

        Re-implements `snowflake.connector.connect()` with connection
        arguments sourced from snowmobile's object model, specifically:
            *   Credentials from `snowmobile.toml`.
            *   Default connection arguments from `snowmobile.toml`.
            *   Optional keyword arguments either passed to
                :meth:`snowmobile.connect()` or directly to this method.

            kwargs:
                Optional keyword arguments to pass to
                snowflake.connector.connect(); arguments passed here will
                over-ride ``connection.default-arguments`` specified in
                ``snowmobile.toml``.

        """
        try:
            self.con = connect(
                **{
                    **self.cfg.connection.connect_kwargs,  # snowmobile.toml
                    **kwargs,  # any kwarg over-rides
                }
            )
            self.cfg._stdout.p(f"..connected: {str(self)}")
            return self

        except DatabaseError as e:
            raise e

    def disconnect(self) -> Snowmobile:
        """Disconnect from connection with which Connection() was instantiated."""
        self.con.close()
        self.con = None
        return self

    @property
    def alive(self) -> bool:
        """Check if the connection is alive."""
        if not isinstance(self.con, SnowflakeConnection):
            return False
        return not self.cursor.is_closed()

    @property
    def cursor(self) -> SnowflakeCursor:
        """:class:`SnowflakeCursor` accessor."""
        if not isinstance(self.con, SnowflakeConnection) and self.ensure_alive:
            self.connect()
        return self.con.cursor()

    # noinspection PydanticTypeChecker,PyTypeChecker
    @property
    def dictcursor(self) -> DictCursor:
        """:class:`DictCursor` accessor."""
        # TODO: check type hint in source code for SnowflakeConnection.cursor()
        #  method to figure out why intellisense is yelling about this;
        #  shouldn't it be Union[SnowflakeCursor, DictCursor]?
        if not isinstance(self.con, SnowflakeConnection) and self.ensure_alive:
            self.connect()
        return self.con.cursor(cursor_class=DictCursor)

    def ex(self, sql: str, on_error: Optional[str] = None, **kwargs) -> SnowflakeCursor:
        """Executes a command via :class:`SnowflakeCursor`.

        Args:
            sql (str):
                ``sql`` command as a string.
            on_error (str):
                String value to impose a specific behavior if an error occurs
                during the execution of ``sql``.
            **kwargs:
                Optional keyword arguments for :meth:`SnowflakeCursor.execute()`.

        Returns (SnowflakeCursor):
            :class:`SnowflakeCursor` object that executed the command.

        """
        try:
            return self.cursor.execute(command=sql, **kwargs)
        except ProgrammingError as e:
            e.to_raise = on_error != 'c'
            self.e.collect(e=e)
            if e.to_raise:
                raise e

    def exd(self, sql: str, on_error: Optional[str] = None, **kwargs) -> DictCursor:
        """Executes a command via :class:`DictCursor`.

        Args:
            sql (str):
                ``sql`` command as a string.
            on_error (str):
                String value to impose a specific behavior if an error occurs
                during the execution of ``sql``.
            **kwargs:
                Optional keyword arguments for :meth:`SnowflakeCursor.execute()`.

        Returns (DictCursor):
            :class:`DictCursor` object that executed the command.

        """
        try:
            return self.dictcursor.execute(command=sql, **kwargs)
        except ProgrammingError as e:
            e.to_raise = on_error != 'c'
            self.e.collect(e=e)
            if e.to_raise:
                raise e

    def query(
        self,
        sql: str,
        as_df: bool = False,
        as_cur: bool = False,
        as_dcur: bool = False,
        as_scalar: bool = False,
        lower: bool = True,
        on_error: Optional[str] = None,
    ) -> Union[pd.DataFrame, SnowflakeCursor]:
        """Execute a query and return results.

         Default behavior of `results=True` will return results as a
         :class:`pandas.DataFrame`, otherwise will execute the sql pr
         with a :class:`SnowflakeCursor` and return the cursor object.

        Args:
            sql (str):
                Raw SQL to execute.
            as_df (bool):
                Return results in DataFrame.
            as_cur (bool):
                Return results in Cursor.
            as_dcur (bool):
                Return results in a DictCursor.
            as_scalar (bool):
                Return results as a single scalar value.
            lower (bool):
                Boolean value indicating whether or not to return results
                with columns lower-cased.
            on_error (str):
                String value to impose a specific behavior if an error occurs
                during the execution of ``sql``.

        Returns (Union[pd.DataFrame, SnowflakeCursor]):
            Results from ``sql`` as a :class:`DataFrame` by default or the
            :class:`SnowflakeCursor` object if `results=False`.

        """
        if not any((as_df, as_cur, as_dcur, as_scalar)):
            as_df = True
        if as_df + as_cur + as_dcur + as_scalar != 1:
            raise ValueError(
                "Only one of ('as_df', 'as_cur', 'as_dcur', 'as_scalar') "
                "can evaluate to `True`"
            )

        if as_cur:
            return self.ex(sql=sql, on_error=on_error)
        if as_dcur:
            return self.exd(sql=sql, on_error=on_error)

        try:
            if not self.alive and self.ensure_alive:
                self.connect()
            df = pd.read_sql(sql, con=self.con)
            if as_df:
                return df.snf.lower() if lower else df
            else:
                return df.snf.to_list(n=1)

        except (pdDataBaseError, DatabaseError) as e:
            e.to_raise = on_error != 'c'
            self.e.collect(e=e)
            if e.to_raise:
                raise e

    def __str__(self) -> str:
        return f"snowmobile.Snowmobile(creds='{self.cfg.connection.creds}')"

    def __repr__(self) -> str:
        return f"snowmobile.Snowmobile(creds='{self.cfg.connection.creds}')"



"""
Inspect the parsing of a sql file containing two bare st.
../docs/snippets/script/a-few-st.py
"""

# Setup -----------------------------------------------------------------------
from pathlib import Path
paths = {p.name: p for p in Path.cwd().glob('**/*.sql')}
path = paths['a-few-tagged-st.sql']

import snowmobile

sn = snowmobile.connect(delay=True)

# Example ---------------------------------------------------------------------

# -- Block 1 --
script = snowmobile.Script(path=path, sn=sn)
script.dtl()
"""
>>>
a-few-tagged-st.sql
===========================
1: Statement('select data~s1')
2: Statement('select data~sample_table')
"""

# -- Block 2 --
s1, s2 = script(1), script(2)
print(s1.nm)  #> select data~s1
print(s2.nm)  #> select data~sample_table

for i, s in script.items():
    print(f"""
{s}
  kw_pr: {s.kw_pr} / kw_ge: {s.kw_ge} / kw: {s.kw}
 obj_pr: {s.obj_pr} / obj_ge: {s.obj_ge} / obj: {s.obj}
desc_pr: {s.desc_pr} / desc_ge: {s.desc_ge} / desc: {s.desc}
  nm_pr: {s.nm_pr} / nm_ge: {s.nm_ge} / nm: {s.nm}
  """)
"""
>>>
Statement('select data~s1')
  kw_pr: '' | kw_ge: select | kw: select
 obj_pr: '' | obj_ge: data | obj: data
desc_pr: '' | desc_ge: s1 | desc: s1
  nm_pr: '' | nm_ge: select data~s1 | nm: select data~s1
  
Statement('select all~sample_table')
  kw_pr: select | kw_ge: select | kw: select
 obj_pr: all | obj_ge: data | obj: all
desc_pr: sample_table | desc_ge: s2 | desc: sample_table
  nm_pr: select all~sample_table | nm_ge: select data~s2 | nm: select all~sample_table
"""


# -- Block 3 --
print(s1.nm_ge)  #> select data~s1
print(s1.nm_pr)  #> None
print(s1.nm)     #> select data~s1

print(s2.nm_ge)  #> select data~s2
print(s2.nm_pr)  #> select data~sample_table
print(s2.nm)     #> select data~sample_table

print(s1.nm_ge)
print(s1.obj_ge)
print(s1.anchor_ge)

print(s2.kw_pr)
print(s2.obj_pr)
print(s2.desc_pr)
print(s2.anchor_pr)
s2.part_desc

anchor_vf = [  # ensure no extra space between words
    sn.cfg.script.power_strip(v, " ")
    for v in s2.part_desc[0].split(" ")
    if sn.cfg.script.power_strip(v, " ")
]
anchor_vf

print(s1.nm_ge)
print(s1.kw)     #> create
print(s1.obj)    #> table
print(s1.index)  #> 1
print(s1.desc)   #> s1
print(s1.nm)     #> create table~s1




"""
Base class for non-configuration objects (i.e. for all classes that are **not**
derived from pydantic's BaseModel).
"""
from __future__ import annotations


class Generic(object):
    """Generic dunder implementation for ``snowmobile`` objects.

    Base class for all ``snowmobile`` objects that do **not** inherit from
    pydantic's BaseModel or configuration class, :class:`Config`.

    """

    def __init__(self):
        pass

    def __getattr__(self, item):
        try:
            return vars(self)[item]
        except KeyError as e:
            raise AttributeError from e

    def __getitem__(self, item):
        return vars(self)[item]

    def __setitem__(self, key, value):
        vars(self)[key] = value

    def __setattr__(self, key, value):
        vars(self)[key] = value

    def __str__(self):
        return f"snowmobile.{type(self).__name__}()"

    def __repr__(self):
        return str(self)




from pathlib import Path
from typing import List
import re
import shutil
import os

MAP_HEADER_SYMBOL = {
	'=': '#',
	'-': '##',
	'~': '###'
}

HERE = Path(__file__).absolute()  # current file location
DOC_ROOT = HERE.parent.parent     # documentation root directory
AUTOAPI_DIR = DOC_ROOT / '_build' / '_sources' / 'autoapi'


class MySTConv:
	"""Converts autoapi rst.txt to compliant MySt."""

	def __init__(self, path: Path):

		self.path = path
		self.path_final = path.parent / f"{path.stem.split('.')[0]}.md.txt"

		self.txt: str = str()

		self.sym1, self.sym2 = '{', '}'
		self.eval_rst = r'{eval-rst}```'

		self.lines = self.txt.split('\n')
		self.lines_vf: List[str] = list()

		self.line_dtl = {}

		self.line_no_f: int = int()

	def get_lines(self):
		"""Identifies lines following headers by header-level."""
		for i, line in enumerate(self.lines):
			if i != 0:
				for s_s, s_t in MAP_HEADER_SYMBOL.items():
					length = len(self.lines[i-1])
					patt = f'{s_s}{self.sym1}{length}{self.sym2}'
					if re.findall(patt, line) and length != 0:
						self.line_dtl[i] = s_s

	def to_myst(self, line: str):
		"""Converts a rst role to an MySt role."""
		if ':' not in line:
			return line
		_, role, obj = line.split(':')
		return f"{self.sym1}{role}{self.sym2}{obj}"

	@property
	def alt_header(self):
		"""First header replacement."""
		self.line_no_f = min(self.line_dtl)
		line_sym = self.line_dtl[self.line_no_f]
		header_sym = MAP_HEADER_SYMBOL[line_sym]
		header_txt = self.to_myst(self.lines[self.line_no_f - 1])
		return f"{header_sym} {header_txt}\n\n{self.eval_rst}\n"

	@property
	def is_swappable(self) -> bool:
		"""Determines if there is an h1 header to be replaced."""
		return min(self.line_dtl) == 1 if len(self.line_dtl) else False

	def swap(self):
		"""Replaced text to save as markdown."""
		self.lines_vf = [line for line in self.lines[self.line_no_f + 2:]]
		self.lines_vf.append('\n```\n')
		self.lines_vf.insert(0, self.alt_header)

	@property
	def txt_final(self) -> str:
		"""Swapped text to saved as markdown."""
		return "\n".join(self.lines_vf)

	def eval(self):
		"""Saves new file as markdown."""
		# read file
		with open(self.path, 'r') as r:
			self.txt = r.read()
		self.get_lines()

		# check if it should be messed with further
		if not self.is_swappable:
			return

		# rename with .md.txt extension and write modified text to it
		self.swap()
		os.remove(path=str(self.path))
		with open(self.path_final, 'w') as f:
			f.write(self.txt_final)
		print(f"\tconverted: {self.path.as_posix()}")


def to_myst():
	"""Batch converts all autoapi RST files to MySt-compliant markdown."""
	files = [
		f for f in AUTOAPI_DIR.rglob('**/*')
		if f.is_file() and str(f).endswith(r'.rst.txt')
	]
	print("\n----/ Starting AutoAPI to MySt /----")
	for f in files:
		c = MySTConv(path=f)
		c.eval()




# import link
from . import link

link_name = "SnowflakeConnection.execute_stream()"
user_text = "SnowflakeConnection.execute_stream()"
url = "https://docs.snowflake.com/en/user-guide/python-connector-example.html#using-execute-stream-to-execute-sql-scripts"

link.xref_links.update({link_name: (user_text, url)})



"""
This file exists for the purpose of exporting DDL.sql to a markdown file.

The relative 'ddl_location' path below assumes that this script is stored
within 'snowmobile/core/pkg_data/.snowmobile' and that `DDL.sql` is in `pkg_data`.
"""
from pathlib import Path

import snowmobile

# location of the DDL.sql file
ddl_location = Path(__file__).absolute().parent.parent / "DDL.sql"

# connector object, connection omitted
sn = snowmobile.connect(delay=True, config_file_nm="snowmobile_testing.toml")

# script object from DDL.sql
script = snowmobile.Script(path=ddl_location, sn=sn)

# accessing as a markup and exporting markdown file only
script.doc().save()



"""
Inspect the parsing of a sql file containing two bare st.
../docs/snippets/script/a-few-st.py
"""

# Setup -----------------------------------------------------------------------
from pathlib import Path
paths = {p.name: p for p in Path.cwd().glob('**/*.sql')}
path = paths['a-few-st.sql']

import snowmobile

sn = snowmobile.connect(delay=True)

# Example ---------------------------------------------------------------------

script = snowmobile.Script(path=path, sn=sn)
script.dtl()
"""
>>>
a-few-statements.sql
====================
1: Statement('create table~s1')
2: Statement('select data~s2')
"""



"""
pydantic Config and Base class for snowmobile.toml and snowmobile-ext.toml
objects.
"""
from __future__ import annotations

import json
from collections.abc import Mapping
from pathlib import Path, WindowsPath
from typing import Any, Callable, Dict, Set

from pydantic import BaseModel, Extra
from pydantic.json import pydantic_encoder


class Config:
    """Configuration class for object model."""

    extra = Extra.allow
    allow_population_by_field_name = True
    allow_population_by_alias = True
    arbitrary_types_allowed = True
    json_encoders = {
        Path: lambda v: str(v.as_posix()),
        WindowsPath: lambda v: str(v.as_posix()),
        type: lambda v: str(v),
        Set: lambda v: list(v),
    }

    def apply_map_by_type(self, attrs: Dict, typ: Any, func: Callable):
        """Recursively apply function to all items of a dictionary of type 'typ'.

        Args:
            attrs (dict):
                Dictionary to traverse.
            typ (Type):
                Type of object to apply function to.
            func (Callable):
                Function to apply to values of type `typ`.

        Returns:
            Altered dictionary with function applied to all keys and values
            matching `typ`.

        """
        if isinstance(attrs, Mapping):
            return {
                self.apply_map_by_type(k, typ, func): self.apply_map_by_type(
                    v, typ, func
                )
                for k, v in attrs.items()
            }
        elif isinstance(attrs, typ):
            return func(attrs)
        else:
            return attrs

    def serialize(self, as_dict: Dict) -> Dict:
        """Recursively applies `json_encoder` functions to all items of a dictionary."""
        serializable = as_dict
        for typ, serialize_func in self.json_encoders.items():
            serializable = self.apply_map_by_type(
                attrs=serializable, typ=typ, func=serialize_func
            )
        return as_dict


class Base(BaseModel, Config):
    """Base class for object model parsed from ``snowmobile.toml``."""

    @property
    def configured_args(self) -> Dict:
        """Placeholder for configuration arguments of derived classes."""
        return dict()

    def kwarg(self, arg_nm: str, arg_val: Any, arg_typ: Any) -> Any:
        """Compares a pr keyword argument to a configured keyword argument."""
        if isinstance(arg_val, arg_typ):
            return arg_val
        return (
            self.configured_args.get(arg_nm)
            if arg_nm in self.configured_args
            else arg_val
        )

    def from_relative(self, obj: Any):
        """Updates current object's attributes with those from a different
        instance of the same class."""
        for k in set(vars(obj)).intersection(set(vars(self))):
            vars(self)[k] = vars(obj)[k]
        return self

    def from_dict(self, args: Dict):
        """Accept a dictionary of arguments and updates the current object as if
        it were instantiated with those arguments."""
        return self.from_relative(obj=type(self)(**args))

    def as_serializable(self, by_alias: bool = False):
        """Returns a dictionary in serializable form."""
        return super().serialize(as_dict=self.dict(by_alias=by_alias))

    def json(self, by_alias: bool = False, **kwargs) -> str:
        """API-facing json serialization method."""
        return self.__json__(by_alias=by_alias, **kwargs)

    def __json__(self, by_alias: bool = False, **kwargs) -> str:
        """Custom serialization to handle sets and pathlib objects."""
        return json.dumps(
            obj=self.as_serializable(by_alias=by_alias),
            default=pydantic_encoder,
            **kwargs,
        )

    def __setattr__(self, key, value):
        vars(self)[key] = value

    def __setitem__(self, key, value):
        vars(self)[key] = value



# import link
from . import link

link_name = "file format"
user_text = "file format"
url = "https://docs.snowflake.com/en/user-guide/data-load-internal-tutorial-create-file-format.html"

link.xref_links.update({link_name: (user_text, url)})



"""
Container for the attributes parsed from a :class:`Tag`.
"""
from __future__ import annotations

from typing import Callable, Dict, Optional, Union

from .connection import Snowmobile


class Attrs(dict):
    """Extended dictionary for attribute storage."""
    
    def __init__(
        self,
        sn: Optional[Snowmobile] = None,
        raw: Optional[str] = None,
        args: Optional[str] = None,
        index: Optional[int] = None,
        **connection_kwargs,
    ):
        
        super().__init__(**(args or dict()))
        
        self.sn: Snowmobile = sn or Snowmobile(
            **{**connection_kwargs, **{'delay': True}}
        )
        """Optional[snowmobile.Snowmobile]: Connection / configuration."""
        
        self.index: int = index
        """int: Index position within the script."""
        
        self._raw: str = raw or str()
        """str: Tag contents as a raw string."""
    
    def _attrs_raw(self, wrap: bool = False) -> str:
        """Prep tag contents based on input."""
        if not wrap or not self._raw:
            return self._raw
        _open, _close = self.sn.cfg.script.tag()
        return f"{_open}{self._raw}{_close}"
    
    @property
    def _attrs_total(self):
        """Parses namespace for attributes specified in **snowmobile.toml**.

        Searches attributes for those matching the keys specified in
        ``script.markdown.attributes.aliases`` within **snowmobile.toml**
        and adds to the existing attributes stored in :attr:`attrs_parsed`
        before returning.

        Returns (dict):
            Combined dictionary of statement attributes from those explicitly
            pr within the script and from object's namespace if specified
            in **snowmobile.toml**.

        """
        current_namespace = {
            **self.sn.cfg.attrs_from_obj(
                obj=self, within=list(self.sn.cfg.attrs.from_namespace)
            ),
            **self.sn.cfg.methods_from_obj(
                obj=self, within=list(self.sn.cfg.attrs.from_namespace)
            ),
        }
        namespace_overlap_with_config = set(current_namespace).intersection(
            self.sn.cfg.attrs.from_namespace  # snowmobile.toml
        )
        attrs = {k: v for k, v in self.items()}  # parsed from .sql
        for k in namespace_overlap_with_config:
            attr = current_namespace[k]
            attr_value = attr() if isinstance(attr, Callable) else attr
            if attr_value:
                attrs[k] = attr_value
        return attrs
    
    def tag(
        self,
        raw: bool = False,
        namespace: bool = False,
        wrap: bool = False,
    ) -> Union[str, Dict, Attrs]:
        """Explicit accessor for self."""
        if raw:
            return self._attrs_raw(wrap=wrap)
        if namespace:
            return self._attrs_total
        return {k: v for k, v in self.items()}
    
    @property
    def is_tagged(self) -> bool:
        """Statement has a prepended tag."""
        return bool(self._raw)
    
    @property
    def is_multiline(self) -> bool:
        """Contains multiline wrap."""
        return '\n' in self._raw



# import link
from . import link

link_name = "DataFrame"
user_text = "DataFrame"
url = "https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.html"

link.xref_links.update({link_name: (user_text, url)})



"""
Inspect the parsing of a sql file.
../docs/snippets/script/intro.py
"""

# Setup -----------------------------------------------------------------------
from pathlib import Path
paths = {p.name: p for p in (Path.cwd() / 'docs' / 'snippets').glob('**/*.sql')}
path = paths['intro.sql']

# Example ---------------------------------------------------------------------

import snowmobile

script = snowmobile.Script(path=path)
markup = script.doc()

print(script)  #> snowmobile.Script('intro.sql')
print(markup)  #> snowmobile.core.Markup('intro.sql')

markup.save()
"""
>>>
    ../intro/intro.md
    ../intro/intro.sql
"""



"""
Create a delayed snowmobile.Snowmobile object.
..docs/snippets/snowmobile/connector_delayed1.py
"""
import snowmobile

sn = snowmobile.connect(delay=True)

type(sn.con)     #> None
print(sn.alive)  #> False

_ = sn.query("select 1")

type(sn.con)     #> snowflake.connector.connection.SnowflakeConnection
print(sn.alive)  #> True

# -- complete example; should run 'as is' --
# snowmobile-include



"""
``snowmobile`` lives in ``snowmobile.core`` to keep from cluttering
intellisense/autocomplete while interacting with the API.

"""
# isort: skip_file
from .base import Generic
from .exception_handler import ExceptionHandler
from .configuration import Configuration
from.connection import Snowmobile
from .connection import Snowmobile as connect
from .section import Section
from .scope import Scope
from .name import Name
from .statement import Statement
from snowmobile.core import errors, cfg, utils
from .column import Column
from .qa import Diff, Empty
from .snowframe import SnowFrame
from .sql import SQL
from .markup import Markup
from .script import Script
from .table import Table
from .paths import (
    # directories
    DIR_MODULES,
    DIR_PKG_DATA,
    # files
    DDL_DEFAULT_PATH,
    EXTENSIONS_DEFAULT_PATH,
    SQL_EXPORT_HEADING_DEFAULT_PATH,
)


__all__ = [
    # core object model
    "Generic",
    "Configuration",
    "Snowmobile",
    "connect",
    "Table",
    "Script",
    "SQL",
    "Scope",
    "Section",
    "Markup",
    "Statement",
    "Diff",
    "Empty",
    "SnowFrame",
    "Column",
    "Name",

    # parsed `snowmobile.toml` objects
    "cfg",

    # error/exception handling
    "ExceptionHandler",
    "errors",

    # file paths
    "DIR_PKG_DATA",
    "DDL_DEFAULT_PATH",
    "EXTENSIONS_DEFAULT_PATH",

    # other
    "utils",
]
